import json
import time
from concurrent import futures
from typing import Generator, Mapping

import grpc
import pytest
from grpc import ServicerContext
from pytest_httpserver import HeaderValueMatcher, HTTPServer
from werkzeug.wrappers import Request, Response

import weaviate
from mock_tests.mock_data import mock_class
from weaviate.connect.base import ConnectionParams, ProtocolParams
from weaviate.proto.v1 import (
    batch_delete_pb2,
    batch_pb2,
    properties_pb2,
    search_get_pb2,
    tenants_pb2,
    weaviate_pb2_grpc,
)

from grpc_health.v1 import health_pb2, health_pb2_grpc

MOCK_IP = "127.0.0.1"
MOCK_PORT = 23536
MOCK_PORT_GRPC = 23537

CLIENT_ID = "DoesNotMatter"
MOCK_SERVER_URL = "http://" + MOCK_IP + ":" + str(MOCK_PORT)
# only http endpoint is tested, grpc port doesnt matter but needs to be supplied
MOCK_SERVER_CONNECTION_PARAMS = ConnectionParams(
    http=ProtocolParams(host=MOCK_IP, port=MOCK_PORT, secure=False),
    grpc=ProtocolParams(host=MOCK_IP, port=MOCK_PORT + 1, secure=False),
)

# pytest_httpserver 'Authorization' HeaderValueMatcher does not work with Bearer tokens.
# Hence, overwrite it with the default header value matcher that just compares for equality.
HeaderValueMatcher.DEFAULT_MATCHERS["Authorization"] = (
    HeaderValueMatcher.default_header_value_matcher
)


@pytest.fixture(scope="session")
def httpserver_listen_address():
    return MOCK_IP, MOCK_PORT


@pytest.fixture(scope="function")
def ready_mock(httpserver: HTTPServer):
    httpserver.expect_request("/v1/.well-known/ready").respond_with_json({})
    yield httpserver


@pytest.fixture(scope="function")
def weaviate_mock(ready_mock: HTTPServer):
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": "1.33"})
    ready_mock.expect_request("/v1/nodes").respond_with_json({"nodes": [{"gitHash": "ABC"}]})

    yield ready_mock


@pytest.fixture(scope="function")
def weaviate_no_auth_mock(weaviate_mock: HTTPServer):
    weaviate_mock.expect_request("/v1/meta").respond_with_json({"version": "1.33"})
    weaviate_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_response(
        Response(json.dumps({}), status=404)
    )

    yield weaviate_mock


@pytest.fixture(scope="function")
def weaviate_auth_mock(weaviate_mock: HTTPServer):
    weaviate_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_json(
        {
            "href": MOCK_SERVER_URL + "/endpoints",
            "clientId": CLIENT_ID,
        }
    )
    weaviate_mock.expect_request("/endpoints").respond_with_json(
        {"token_endpoint": MOCK_SERVER_URL + "/auth"}
    )
    yield weaviate_mock


@pytest.fixture(scope="function")
def weaviate_timeouts_mock(weaviate_no_auth_mock: HTTPServer):
    def slow_get(request: Request) -> Response:
        time.sleep(1)
        return Response(json.dumps({"doesn't": "matter"}), content_type="application/json")

    def slow_post(request: Request) -> Response:
        time.sleep(2)
        return Response(json.dumps({"doesn't": "matter"}), content_type="application/json")

    weaviate_no_auth_mock.expect_request(
        f"/v1/schema/{mock_class['class']}", method="GET"
    ).respond_with_handler(slow_get)
    weaviate_no_auth_mock.expect_request("/v1/objects", method="POST").respond_with_handler(
        slow_post
    )

    yield weaviate_no_auth_mock


@pytest.fixture(scope="function")
def start_grpc_server() -> Generator[grpc.Server, None, None]:
    # Create a gRPC server
    server: grpc.Server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

    # Implement the health check service
    class MockHealthServicer(health_pb2_grpc.HealthServicer):
        def Check(
            self, request: health_pb2.HealthCheckRequest, context: ServicerContext
        ) -> health_pb2.HealthCheckResponse:
            return health_pb2.HealthCheckResponse(status=health_pb2.HealthCheckResponse.SERVING)

    # Add the health check service to the server
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)

    # Listen on a specific port
    server.add_insecure_port(f"[::]:{MOCK_PORT_GRPC}")
    server.start()

    yield server

    # Teardown - stop the server
    server.stop(0)


@pytest.fixture(scope="function")
def weaviate_client(
    weaviate_mock: HTTPServer, start_grpc_server: grpc.Server
) -> Generator[weaviate.WeaviateClient, None, None]:
    client = weaviate.connect_to_local(port=MOCK_PORT, host=MOCK_IP, grpc_port=MOCK_PORT_GRPC)
    yield client
    client.close()


@pytest.fixture(scope="function")
def weaviate_timeouts_client(
    weaviate_timeouts_mock: HTTPServer, start_grpc_server: grpc.Server
) -> Generator[weaviate.WeaviateClient, None, None]:
    client = weaviate.connect_to_local(
        host=MOCK_IP,
        port=MOCK_PORT,
        grpc_port=MOCK_PORT_GRPC,
        additional_config=weaviate.classes.init.AdditionalConfig(
            timeout=weaviate.classes.init.Timeout(query=0.5, insert=1.5)
        ),
    )
    yield client
    client.close()


@pytest.fixture(scope="function")
def tenants_collection(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    class MockWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
        def TenantsGet(
            self, request: tenants_pb2.TenantsGetRequest, context: ServicerContext
        ) -> tenants_pb2.TenantsGetReply:
            return tenants_pb2.TenantsGetReply(
                tenants=[
                    tenants_pb2.Tenant(
                        name="tenant1", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_HOT
                    ),
                    tenants_pb2.Tenant(
                        name="tenant2", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_COLD
                    ),
                    tenants_pb2.Tenant(
                        name="tenant3", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_FROZEN
                    ),
                    tenants_pb2.Tenant(
                        name="tenant4", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_FREEZING
                    ),
                    tenants_pb2.Tenant(
                        name="tenant5",
                        activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_UNFREEZING,
                    ),
                    tenants_pb2.Tenant(
                        name="tenant6", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_ACTIVE
                    ),
                    tenants_pb2.Tenant(
                        name="tenant7", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_INACTIVE
                    ),
                    tenants_pb2.Tenant(
                        name="tenant8", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_OFFLOADED
                    ),
                    tenants_pb2.Tenant(
                        name="tenant9",
                        activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_OFFLOADING,
                    ),
                    tenants_pb2.Tenant(
                        name="tenant10",
                        activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_ONLOADING,
                    ),
                ]
            )

    weaviate_pb2_grpc.add_WeaviateServicer_to_server(MockWeaviateService(), start_grpc_server)
    return weaviate_client.collections.use("TenantsGetCollectionName")


@pytest.fixture(scope="function")
def year_zero_collection(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    class MockWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
        def Search(
            self, request: search_get_pb2.SearchRequest, context: grpc.ServicerContext
        ) -> search_get_pb2.SearchReply:
            zero_date = properties_pb2.Value(date_value="0000-01-30T00:00:00Z")
            date_prop: Mapping[str, properties_pb2.Value] = {"date": zero_date}
            return search_get_pb2.SearchReply(
                results=[
                    search_get_pb2.SearchResult(
                        properties=search_get_pb2.PropertiesResult(
                            non_ref_props=properties_pb2.Properties(fields=date_prop)
                        )
                    ),
                ]
            )

    weaviate_pb2_grpc.add_WeaviateServicer_to_server(MockWeaviateService(), start_grpc_server)
    return weaviate_client.collections.use("YearZeroCollection")


@pytest.fixture(scope="function")
def timeouts_collection(
    weaviate_timeouts_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    class MockWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
        def Search(
            self, request: search_get_pb2.SearchRequest, context: grpc.ServicerContext
        ) -> search_get_pb2.SearchReply:
            time.sleep(1)
            return search_get_pb2.SearchReply()

        def BatchObjects(
            self, request: batch_pb2.BatchObjectsRequest, context: grpc.ServicerContext
        ) -> batch_pb2.BatchObjectsReply:
            time.sleep(2)
            return batch_pb2.BatchObjectsReply()

    weaviate_pb2_grpc.add_WeaviateServicer_to_server(MockWeaviateService(), start_grpc_server)
    return weaviate_timeouts_client.collections.use(mock_class["class"])


class MockRetriesWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
    search_count = 0
    tenants_count = 0

    def Search(
        self, request: search_get_pb2.SearchRequest, context: grpc.ServicerContext
    ) -> search_get_pb2.SearchReply:
        if self.search_count == 0:
            self.search_count += 1
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details("Internal server error")
            return search_get_pb2.SearchReply()
        if self.search_count == 1:
            self.search_count += 1
            context.set_code(grpc.StatusCode.UNAVAILABLE)
            context.set_details("Service is unavailable")
            return search_get_pb2.SearchReply()
        return search_get_pb2.SearchReply(
            results=[
                search_get_pb2.SearchResult(
                    properties=search_get_pb2.PropertiesResult(
                        non_ref_props=properties_pb2.Properties(
                            fields={"name": properties_pb2.Value(text_value="test")}
                        )
                    )
                )
            ]
        )

    def TenantsGet(
        self, request: tenants_pb2.TenantsGetRequest, context: ServicerContext
    ) -> tenants_pb2.TenantsGetReply:
        if self.tenants_count == 0:
            self.tenants_count += 1
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details("Internal server error")
            return tenants_pb2.TenantsGetReply()
        if self.tenants_count == 1:
            self.tenants_count += 1
            context.set_code(grpc.StatusCode.UNAVAILABLE)
            context.set_details("Service is unavailable")
            return tenants_pb2.TenantsGetReply()
        return tenants_pb2.TenantsGetReply(
            tenants=[
                tenants_pb2.Tenant(
                    name="tenant1", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_ACTIVE
                )
            ]
        )


@pytest.fixture(scope="function")
def retries(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> tuple[weaviate.collections.Collection, MockRetriesWeaviateService]:
    service = MockRetriesWeaviateService()
    weaviate_pb2_grpc.add_WeaviateServicer_to_server(service, start_grpc_server)
    return weaviate_client.collections.use("RetriesCollection"), service


class MockForbiddenWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
    def Search(
        self, request: search_get_pb2.SearchRequest, context: grpc.ServicerContext
    ) -> search_get_pb2.SearchReply:
        context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        context.set_details("Permission denied")
        return search_get_pb2.SearchReply()

    def TenantsGet(
        self, request: tenants_pb2.TenantsGetRequest, context: ServicerContext
    ) -> tenants_pb2.TenantsGetReply:
        context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        context.set_details("Permission denied")
        return tenants_pb2.TenantsGetReply()

    def BatchObjects(
        self, request: batch_pb2.BatchObjectsRequest, context: grpc.ServicerContext
    ) -> batch_pb2.BatchObjectsReply:
        context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        context.set_details("Permission denied")
        return batch_pb2.BatchObjectsReply()

    def BatchDelete(
        self, request: batch_delete_pb2.BatchDeleteRequest, context: grpc.ServicerContext
    ) -> batch_delete_pb2.BatchDeleteReply:
        context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        context.set_details("Permission denied")
        return batch_delete_pb2.BatchDeleteReply()


@pytest.fixture(scope="function")
def forbidden(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    service = MockForbiddenWeaviateService()
    weaviate_pb2_grpc.add_WeaviateServicer_to_server(service, start_grpc_server)
    return weaviate_client.collections.use("ForbiddenCollection")



import pytest

from integration.conftest import ClientFactory


@pytest.fixture
def sorting_test_client(client_factory: ClientFactory):
    client = client_factory()
    try:
        yield client
    finally:
        client.close()


def test_collections_list_all_sorting(sorting_test_client):
    """Test that collections.list_all() returns collections sorted alphabetically by key."""
    client = sorting_test_client

    try:
        # Create collections with names in non-alphabetical order
        client.collections.create(name="SortingTestCollectionC")
        client.collections.create(name="SortingTestCollectionA")
        client.collections.create(name="SortingTestCollectionB")

        # Get all collections
        collections = client.collections.list_all()

        # Get the keys and filter only our test collections
        collection_keys = list(collections.keys())
        test_collections = [k for k in collection_keys if k.startswith("SortingTestCollection")]

        # Verify they are in alphabetical order
        assert test_collections == sorted(test_collections)

        # Test with simple=False as well
        collections = client.collections.list_all(simple=False)
        collection_keys = list(collections.keys())
        test_collections = [k for k in collection_keys if k.startswith("SortingTestCollection")]
        assert test_collections == sorted(test_collections)

    finally:
        # Clean up
        client.collections.delete(
            ["SortingTestCollectionA", "SortingTestCollectionB", "SortingTestCollectionC"]
        )



from .generate import _HybridGenerate, _HybridGenerateAsync
from .query import _HybridQuery, _HybridQueryAsync

__all__ = [
    "_HybridGenerate",
    "_HybridGenerateAsync",
    "_HybridQuery",
    "_HybridQueryAsync",
]



import pytest

import weaviate
from weaviate.exceptions import WeaviateQueryError, WeaviateTimeoutError


def test_timeout_rest_query(timeouts_collection: weaviate.collections.Collection):
    with pytest.raises(WeaviateTimeoutError):
        timeouts_collection.config.get()


def test_timeout_rest_insert(timeouts_collection: weaviate.collections.Collection):
    with pytest.raises(WeaviateTimeoutError):
        timeouts_collection.data.insert(properties={"what": "ever"})


def test_timeout_grpc_query(timeouts_collection: weaviate.collections.Collection):
    with pytest.raises(WeaviateQueryError) as recwarn:
        timeouts_collection.query.fetch_objects()
        assert "DEADLINE_EXCEEDED" in str(recwarn)


def test_timeout_grpc_insert(timeouts_collection: weaviate.collections.Collection):
    with pytest.raises(WeaviateQueryError) as recwarn:
        timeouts_collection.data.insert_many([{"what": "ever"}])
        assert "DEADLINE_EXCEEDED" in str(recwarn)



import hashlib
import os
import platform
import re
import socket
import stat
import subprocess
import tarfile
import time
import urllib.request
import warnings
import zipfile
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple

import httpx
import validators

from weaviate import exceptions
from weaviate.exceptions import WeaviateStartUpError
from weaviate.logger import logger
from weaviate.util import _decode_json_response_dict

DEFAULT_BINARY_PATH = str(Path.home() / ".cache/weaviate-embedded/")
DEFAULT_PERSISTENCE_DATA_PATH = str(Path.home() / ".local/share/weaviate")
GITHUB_RELEASE_DOWNLOAD_URL = "https://github.com/weaviate/weaviate/releases/download/"

DEFAULT_PORT = 8079
DEFAULT_GRPC_PORT = 50060

WEAVIATE_VERSION = "1.30.5"


@dataclass
class EmbeddedOptions:
    persistence_data_path: str = os.environ.get("XDG_DATA_HOME", DEFAULT_PERSISTENCE_DATA_PATH)
    binary_path: str = os.environ.get("XDG_CACHE_HOME", DEFAULT_BINARY_PATH)
    version: str = WEAVIATE_VERSION
    port: int = DEFAULT_PORT
    hostname: str = "127.0.0.1"
    additional_env_vars: Optional[Dict[str, str]] = None
    grpc_port: int = DEFAULT_GRPC_PORT


def get_random_port() -> int:
    sock = socket.socket()
    sock.bind(("", 0))
    port_num = int(sock.getsockname()[1])
    sock.close()
    return port_num


class _EmbeddedBase:
    def __init__(self, options: EmbeddedOptions) -> None:
        self.options = options
        self.grpc_port: int = options.grpc_port
        self.process: Optional[subprocess.Popen[bytes]] = None
        self.ensure_paths_exist()
        self.check_supported_platform()
        self._parsed_weaviate_version = ""
        # regular expression to detect a version number: v[one digit].[1-2 digits].[1-2 digits]
        # optionally there can be a "-rc/alpha/beta.[1-2 digits]"
        # nothing in front or back
        version_pattern = re.compile(
            r"^\d\.\d{1,2}\.\d{1,2}?(-rc\.\d{1,2}|-beta\.\d{1,2}|-alpha\.\d{1,2}|$)$"
        )

        valid_url = validators.url(self.options.version)
        if isinstance(valid_url, validators.ValidationError):
            valid_url = validators.url(self.options.version, simple_host=True)  # for localhost
        if valid_url:
            if not self.options.version.endswith(".tar.gz") and not self.options.version.endswith(
                ".zip"
            ):
                raise exceptions.WeaviateEmbeddedInvalidVersionError(self.options.version)

            # for GitHub urls we can parse the version from the url
            if self.options.version.startswith(GITHUB_RELEASE_DOWNLOAD_URL):
                # replace with str.removeprefix() after 3.8 has been deprecated
                self._parsed_weaviate_version = self.options.version[
                    len(GITHUB_RELEASE_DOWNLOAD_URL) :
                ].split("/")[0]
            self._download_url = self.options.version
        elif version_pattern.match(self.options.version):
            version_tag = "v" + self.options.version
            self._parsed_weaviate_version = version_tag
            self._set_download_url_from_version_tag(version_tag)
        elif self.options.version == "latest":
            response = httpx.get("https://api.github.com/repos/weaviate/weaviate/releases/latest")
            latest = _decode_json_response_dict(response, "get tag of latest weaviate release")
            assert latest is not None
            version_tag = latest["tag_name"]
            self._parsed_weaviate_version = version_tag
            self._set_download_url_from_version_tag(version_tag)
        else:
            raise exceptions.WeaviateEmbeddedInvalidVersionError(self.options.version)

    def _set_download_url_from_version_tag(self, version: str) -> None:
        if platform.system() == "Darwin":
            machine_type = "all"
            package_format = "zip"
        else:
            machine_type = platform.machine()
            if machine_type == "x86_64":
                machine_type = "amd64"
            elif machine_type == "aarch64":
                machine_type = "arm64"
            package_format = "tar.gz"

        self._download_url = (
            GITHUB_RELEASE_DOWNLOAD_URL
            + version
            + "/weaviate-"
            + version
            + "-"
            + platform.system()
            + "-"
            + machine_type
            + "."
            + package_format
        )

    def __del__(self) -> None:
        self.stop()

    def ensure_paths_exist(self) -> None:
        Path(self.options.binary_path).mkdir(parents=True, exist_ok=True)
        Path(self.options.persistence_data_path).mkdir(parents=True, exist_ok=True)

    def ensure_weaviate_binary_exists(self) -> None:
        self._weaviate_binary_path = Path(
            self.options.binary_path,
            "weaviate-"
            + self._parsed_weaviate_version
            + "-"
            + str(hashlib.sha256(self._parsed_weaviate_version.encode("utf-8")).hexdigest()),
        )
        if not self._weaviate_binary_path.exists():
            logger.info(
                f"Binary {self.options.binary_path} did not exist. Downloading binary from {self._download_url}"
            )
            if self._download_url.endswith(".tar.gz"):
                tar_filename = Path(self.options.binary_path, "tmp_weaviate.tgz")
                urllib.request.urlretrieve(self._download_url, tar_filename)
                with tarfile.open(tar_filename) as binary_tar:
                    binary_tar.extract("weaviate", path=Path(self.options.binary_path))
                tar_filename.unlink()
            else:
                assert self._download_url.endswith(".zip")
                zip_filename = Path(self.options.binary_path, "tmp_weaviate.zip")
                urllib.request.urlretrieve(self._download_url, zip_filename)
                with zipfile.ZipFile(zip_filename, "r") as zip_ref:
                    zip_ref.extract("weaviate", path=Path(self.options.binary_path))

            (Path(self.options.binary_path) / "weaviate").rename(self._weaviate_binary_path)

            # Ensuring weaviate binary is executable
            self._weaviate_binary_path.chmod(
                self._weaviate_binary_path.stat().st_mode | stat.S_IEXEC
            )

    def wait_till_listening(self) -> None:
        seconds = 30
        sleep_interval = 0.1
        retries = int(seconds / sleep_interval)
        while self.is_listening() is False and retries > 0:
            time.sleep(sleep_interval)
            retries -= 1
        if retries == 0:
            raise WeaviateStartUpError(
                f"Embedded DB did not start listening on port {self.options.port} within {seconds} seconds"
            )

    @staticmethod
    def check_supported_platform() -> None:
        if platform.system() in ["Windows"]:
            raise WeaviateStartUpError(
                f"""{platform.system()} is not supported with EmbeddedDB. Please upvote this feature request if you want
                 this: https://github.com/weaviate/weaviate/issues/3315"""  # noqa: E231
            )

    def stop(self) -> None:
        if self.process is not None:
            try:
                self.process.terminate()
                self.process.wait()
            except ProcessLookupError:
                logger.info(
                    f"""Tried to stop embedded weaviate process {self.process.pid}. Process was not found. So not doing
                    anything"""
                )
            self.process = None

    def ensure_running(self) -> None:
        if self.is_listening() is False:
            logger.info(
                f"Embedded weaviate wasn't listening on ports http:{self.options.port} & grpc:{self.options.grpc_port}, so starting embedded weaviate again"
            )
            self.start()

    def start(self) -> None:
        self.ensure_weaviate_binary_exists()
        my_env = os.environ.copy()

        my_env.setdefault("AUTHENTICATION_ANONYMOUS_ACCESS_ENABLED", "true")
        my_env.setdefault("QUERY_DEFAULTS_LIMIT", "20")
        my_env.setdefault("PERSISTENCE_DATA_PATH", self.options.persistence_data_path)
        my_env.setdefault("PROFILING_PORT", str(get_random_port()))
        # Limitation with weaviate server requires setting
        # data_bind_port to gossip_bind_port + 1
        gossip_bind_port = get_random_port()
        data_bind_port = gossip_bind_port + 1
        my_env.setdefault("CLUSTER_GOSSIP_BIND_PORT", str(gossip_bind_port))
        my_env.setdefault("CLUSTER_DATA_BIND_PORT", str(data_bind_port))
        my_env.setdefault("GRPC_PORT", str(self.grpc_port))
        my_env.setdefault("RAFT_BOOTSTRAP_EXPECT", str(1))
        my_env.setdefault("CLUSTER_IN_LOCALHOST", str(True))

        # Each call to `get_random_port()` will likely result in
        # a port 1 higher than the last time it was called. With
        # this, we end up with raft_port == gossip_bind_port + 1,
        # which is the same as data_bind_port. This kind of
        # configuration leads to failed cross cluster communication.
        # Although the current version of embedded does not support
        # multi-node instances, the backup process communication
        # passes through the internal cluster server, and will fail.
        #
        # So we here we ensure that raft_port never collides with
        # data_bind_port.
        raft_port = data_bind_port + 1
        raft_internal_rpc_port = raft_port + 1
        my_env.setdefault("RAFT_PORT", str(raft_port))
        my_env.setdefault("RAFT_INTERNAL_RPC_PORT", str(raft_internal_rpc_port))

        my_env.setdefault(
            "ENABLE_MODULES",
            "text2vec-openai,text2vec-cohere,text2vec-huggingface,ref2vec-centroid,generative-openai,qna-openai,"
            "reranker-cohere",
        )

        # have a deterministic hostname in case of changes in the network name.
        # This allows to run multiple parallel instances
        cluster_hostname = f"Embedded_at_{self.options.port}"
        my_env.setdefault("CLUSTER_HOSTNAME", cluster_hostname)
        my_env.setdefault("RAFT_JOIN", f"{cluster_hostname}:{raft_port}")

        if self.options.additional_env_vars is not None:
            my_env.update(self.options.additional_env_vars)

        # filter warning about running processes.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            process = subprocess.Popen(
                [
                    f"{self._weaviate_binary_path}",
                    "--host",
                    self.options.hostname,
                    "--port",
                    str(self.options.port),
                    "--scheme",
                    "http",
                    "--read-timeout=600s",
                    "--write-timeout=600s",
                ],
                env=my_env,
            )
            self.process = process
        logger.info(f"Started {self.options.binary_path}: process ID {self.process.pid}")
        self.wait_till_listening()

    @abstractmethod
    def is_listening(self) -> bool:
        raise NotImplementedError()


class EmbeddedV3(_EmbeddedBase):
    def is_listening(self) -> bool:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.options.hostname, self.options.port))
            return True
        except (socket.error, ConnectionRefusedError):
            return False
        finally:
            s.close()

    def start(self) -> None:
        if self.is_listening():
            logger.info(f"embedded weaviate is already listening on port {self.options.port}")
            return
        super().start()


EmbeddedDB = EmbeddedV3  # needed for BC from v3 -> v4


class EmbeddedV4(_EmbeddedBase):
    def is_listening(self) -> bool:
        up = self.__is_listening()
        return up[0] and up[1]

    def __is_listening(self) -> Tuple[bool, bool]:
        http_listening, grpc_listening = False, False
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect((self.options.hostname, self.options.port))
                http_listening = True
            except (socket.error, ConnectionRefusedError):
                pass
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect((self.options.hostname, self.grpc_port))
                grpc_listening = True
            except (socket.error, ConnectionRefusedError):
                pass
        return (http_listening, grpc_listening)

    def start(self) -> None:
        up = self.__is_listening()
        if up[0] and up[1]:
            raise WeaviateStartUpError(
                f"Embedded DB did not start because processes are already listening on ports http:{self.options.port} and grpc:{self.grpc_port}"
                f"use weaviate.connect_to_local(port={self.options.port}, grpc_port={self.options.grpc_port}) to connect to the existing instance"
            )
        elif up[0] and not up[1]:
            raise WeaviateStartUpError(
                f"Embedded DB did not start because a process is already listening on port http:{self.options.port}"
                "look for another free port for the HTTP connection to you embedded instance"
            )
        elif up[1] and not up[0]:
            raise WeaviateStartUpError(
                f"Embedded DB did not start because a process is already listening on port grpc:{self.grpc_port}"
                "look for another free port for the gRPC connection to your embedded instance"
            )
        super().start()



from typing import Optional, Type

from weaviate.collections.classes.config import CollectionConfig
from weaviate.collections.classes.internal import Properties, References
from weaviate.collections.collection.sync import Collection
from weaviate.collections.collections.base import _CollectionsBase
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _Collections(_CollectionsBase[ConnectionSync]):
    @executor.no_wrapping
    def use(
        self,
        name: str,
        data_model_properties: Optional[Type[Properties]] = None,
        data_model_references: Optional[Type[References]] = None,
        skip_argument_validation: bool = False,
    ) -> Collection[Properties, References]:
        collection = self._use(
            name=name,
            data_model_properties=data_model_properties,
            data_model_references=data_model_references,
            skip_argument_validation=skip_argument_validation,
        )
        assert isinstance(collection, Collection)
        return collection

    @executor.no_wrapping
    def create_from_dict(self, config: dict) -> Collection:
        collection = executor.result(self._create_from_dict(config))
        assert isinstance(collection, Collection)
        return collection

    @executor.no_wrapping
    def create_from_config(self, config: CollectionConfig) -> Collection:
        collection = executor.result(self._create_from_config(config))
        assert isinstance(collection, Collection)
        return collection



from typing import Dict, Optional, Sequence, Union

from weaviate.collections.tenants.types import (
    TenantCreateInputType,
    TenantInputType,
    TenantOutputType,
    TenantUpdateInputType,
)
from weaviate.connect.v4 import ConnectionAsync

from .executor import _TenantsExecutor

class _TenantsAsync(_TenantsExecutor[ConnectionAsync]):
    async def create(
        self, tenants: Union[TenantCreateInputType, Sequence[TenantCreateInputType]]
    ) -> None: ...
    async def remove(self, tenants: Union[TenantInputType, Sequence[TenantInputType]]) -> None: ...
    async def get(self) -> Dict[str, TenantOutputType]: ...
    async def get_by_names(
        self, tenants: Sequence[TenantInputType]
    ) -> Dict[str, TenantOutputType]: ...
    async def get_by_name(self, tenant: TenantInputType) -> Optional[TenantOutputType]: ...
    async def update(
        self, tenants: Union[TenantUpdateInputType, Sequence[TenantUpdateInputType]]
    ) -> None: ...
    async def exists(self, tenant: TenantInputType) -> bool: ...
    async def activate(self, tenant: Union[TenantInputType, Sequence[TenantInputType]]) -> None: ...
    async def deactivate(
        self, tenant: Union[TenantInputType, Sequence[TenantInputType]]
    ) -> None: ...
    async def offload(self, tenant: Union[TenantInputType, Sequence[TenantInputType]]) -> None: ...



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    TypedDict,
    TypeVar,
    Union,
)

from pydantic import BaseModel
from typing_extensions import NotRequired

from weaviate.cluster.types import Verbosity
from weaviate.str_enum import BaseEnum
from weaviate.util import _capitalize_first_letter
from weaviate.warnings import _Warnings


class UserTypes(str, Enum):
    DB_DYNAMIC = "db_user"
    DB_STATIC = "db_env_user"
    OIDC = "oidc"


class GroupTypes(str, Enum):
    OIDC = "oidc"


@dataclass
class UserAssignment:
    user_id: str
    user_type: UserTypes


@dataclass
class GroupAssignment:
    group_id: str
    group_type: GroupTypes


class WeaviateUserAssignment(TypedDict):
    userId: str
    userType: str


class RoleScope(str, BaseEnum):
    """Scope of the role permission."""

    MATCH = "match"
    ALL = "all"


class PermissionData(TypedDict):
    collection: str
    tenant: str


class PermissionCollections(TypedDict):
    collection: str


class PermissionsTenants(TypedDict):
    collection: str
    tenant: str


class PermissionNodes(TypedDict):
    collection: str
    verbosity: Verbosity


class PermissionBackup(TypedDict):
    collection: str


class PermissionReplicate(TypedDict):
    collection: str
    shard: str


class PermissionRoles(TypedDict):
    role: str
    scope: NotRequired[str]


class PermissionsUsers(TypedDict):
    users: str


class PermissionsGroups(TypedDict):
    group: str
    groupType: str


class PermissionsAlias(TypedDict):
    alias: str
    collection: str


# action is always present in WeaviatePermission
class WeaviatePermissionRequired(TypedDict):
    action: str


class WeaviatePermission(
    WeaviatePermissionRequired, total=False
):  # Add total=False to make all fields optional
    data: Optional[PermissionData]
    collections: Optional[PermissionCollections]
    nodes: Optional[PermissionNodes]
    backups: Optional[PermissionBackup]
    replicate: Optional[PermissionReplicate]
    roles: Optional[PermissionRoles]
    tenants: Optional[PermissionsTenants]
    users: Optional[PermissionsUsers]
    aliases: Optional[PermissionsAlias]
    groups: Optional[PermissionsGroups]


class WeaviateRole(TypedDict):
    name: str
    permissions: List[WeaviatePermission]


class WeaviateUser(TypedDict):
    user_id: str
    roles: Optional[List[WeaviateRole]]
    groups: List[str]


class WeaviateDBUserRoleNames(TypedDict):
    userId: str
    roles: List[str]
    groups: List[str]
    active: bool
    dbUserType: str


class _Action:
    pass


class AliasAction(str, _Action, Enum):
    CREATE = "create_aliases"
    READ = "read_aliases"
    UPDATE = "update_aliases"
    DELETE = "delete_aliases"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in AliasAction]


class CollectionsAction(str, _Action, Enum):
    CREATE = "create_collections"
    READ = "read_collections"
    UPDATE = "update_collections"
    DELETE = "delete_collections"
    MANAGE = "manage_collections"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in CollectionsAction]


class TenantsAction(str, _Action, Enum):
    CREATE = "create_tenants"
    READ = "read_tenants"
    UPDATE = "update_tenants"
    DELETE = "delete_tenants"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in TenantsAction]


class DataAction(str, _Action, Enum):
    CREATE = "create_data"
    READ = "read_data"
    UPDATE = "update_data"
    DELETE = "delete_data"
    MANAGE = "manage_data"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in DataAction]


class RolesAction(str, _Action, Enum):
    MANAGE = "manage_roles"  # backward compatibility, remove in a bit
    CREATE = "create_roles"
    READ = "read_roles"
    UPDATE = "update_roles"
    DELETE = "delete_roles"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in RolesAction]


class GroupAction(str, _Action, Enum):
    READ = "read_groups"
    ASSIGN_AND_REVOKE = "assign_and_revoke_groups"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in GroupAction]


class UsersAction(str, _Action, Enum):
    CREATE = "create_users"
    READ = "read_users"
    UPDATE = "update_users"
    DELETE = "delete_users"
    ASSIGN_AND_REVOKE = "assign_and_revoke_users"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in UsersAction]


class ClusterAction(str, _Action, Enum):
    READ = "read_cluster"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in ClusterAction]


class NodesAction(str, _Action, Enum):
    READ = "read_nodes"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in NodesAction]


class BackupsAction(str, _Action, Enum):
    MANAGE = "manage_backups"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in BackupsAction]


class ReplicateAction(str, _Action, Enum):
    CREATE = "create_replicate"
    READ = "read_replicate"
    UPDATE = "update_replicate"
    DELETE = "delete_replicate"

    @staticmethod
    def values() -> List[str]:
        return [action.value for action in ReplicateAction]


ActionT = TypeVar("ActionT", bound=Enum)


class _Permission(BaseModel, Generic[ActionT]):
    actions: Set[ActionT]

    @abstractmethod
    def _to_weaviate(self) -> List[WeaviatePermission]:
        raise NotImplementedError()


class _CollectionsPermission(_Permission[CollectionsAction]):
    collection: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "collections": {
                    "collection": _capitalize_first_letter(self.collection),
                },
            }
            for action in self.actions
        ]


class _TenantsPermission(_Permission[TenantsAction]):
    collection: str
    tenant: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "tenants": {
                    "collection": _capitalize_first_letter(self.collection),
                    "tenant": self.tenant,
                },
            }
            for action in self.actions
        ]


class _NodesPermission(_Permission[NodesAction]):
    verbosity: Verbosity
    collection: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "nodes": {
                    "collection": _capitalize_first_letter(self.collection),
                    "verbosity": self.verbosity,
                },
            }
            for action in self.actions
        ]


class _ReplicatePermission(_Permission[ReplicateAction]):
    collection: str
    shard: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "replicate": {
                    "collection": _capitalize_first_letter(self.collection),
                    "shard": self.shard,
                },
            }
            for action in self.actions
        ]


class _RolesPermission(_Permission[RolesAction]):
    role: str
    scope: Optional[str] = None

    def _to_weaviate(self) -> List[WeaviatePermission]:
        roles: PermissionRoles = {"role": self.role}
        if self.scope is not None:
            roles["scope"] = self.scope
        return [
            {
                "action": action,
                "roles": roles,
            }
            for action in self.actions
        ]


class _UsersPermission(_Permission[UsersAction]):
    users: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [{"action": action, "users": {"users": self.users}} for action in self.actions]


class _AliasPermission(_Permission[AliasAction]):
    alias: str
    collection: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "aliases": {
                    "alias": _capitalize_first_letter(self.alias),
                    "collection": self.collection,
                },
            }
            for action in self.actions
        ]


class _GroupsPermission(_Permission[GroupAction]):
    group: str
    group_type: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {"action": action, "groups": {"group": self.group, "groupType": self.group_type}}
            for action in self.actions
        ]


class _BackupsPermission(_Permission[BackupsAction]):
    collection: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "backups": {
                    "collection": _capitalize_first_letter(self.collection),
                },
            }
            for action in self.actions
        ]


class _ClusterPermission(_Permission[ClusterAction]):
    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
            }
            for action in self.actions
        ]


class _DataPermission(_Permission[DataAction]):
    collection: str
    tenant: str

    def _to_weaviate(self) -> List[WeaviatePermission]:
        return [
            {
                "action": action,
                "data": {
                    "collection": _capitalize_first_letter(self.collection),
                    "tenant": self.tenant,
                },
            }
            for action in self.actions
        ]


class CollectionsPermissionOutput(_CollectionsPermission):
    pass


class DataPermissionOutput(_DataPermission):
    pass


class ReplicatePermissionOutput(_ReplicatePermission):
    pass


class RolesPermissionOutput(_RolesPermission):
    pass


class UsersPermissionOutput(_UsersPermission):
    pass


class GroupsPermissionOutput(_GroupsPermission):
    pass


class AliasPermissionOutput(_AliasPermission):
    pass


class ClusterPermissionOutput(_ClusterPermission):
    pass


class BackupsPermissionOutput(_BackupsPermission):
    pass


class NodesPermissionOutput(_NodesPermission):
    pass


class TenantsPermissionOutput(_TenantsPermission):
    pass


PermissionsOutputType = Union[
    AliasPermissionOutput,
    ClusterPermissionOutput,
    CollectionsPermissionOutput,
    DataPermissionOutput,
    RolesPermissionOutput,
    UsersPermissionOutput,
    BackupsPermissionOutput,
    NodesPermissionOutput,
    TenantsPermissionOutput,
    ReplicatePermissionOutput,
    GroupsPermissionOutput,
]


@dataclass
class RoleBase:
    name: str


@dataclass
class Role(RoleBase):
    alias_permissions: List[AliasPermissionOutput]
    cluster_permissions: List[ClusterPermissionOutput]
    collections_permissions: List[CollectionsPermissionOutput]
    data_permissions: List[DataPermissionOutput]
    roles_permissions: List[RolesPermissionOutput]
    users_permissions: List[UsersPermissionOutput]
    backups_permissions: List[BackupsPermissionOutput]
    nodes_permissions: List[NodesPermissionOutput]
    tenants_permissions: List[TenantsPermissionOutput]
    replicate_permissions: List[ReplicatePermissionOutput]
    groups_permissions: List[GroupsPermissionOutput]

    @property
    def permissions(self) -> List[PermissionsOutputType]:
        permissions: List[PermissionsOutputType] = []
        permissions.extend(self.alias_permissions)
        permissions.extend(self.cluster_permissions)
        permissions.extend(self.collections_permissions)
        permissions.extend(self.data_permissions)
        permissions.extend(self.roles_permissions)
        permissions.extend(self.users_permissions)
        permissions.extend(self.backups_permissions)
        permissions.extend(self.nodes_permissions)
        permissions.extend(self.tenants_permissions)
        permissions.extend(self.replicate_permissions)
        permissions.extend(self.groups_permissions)
        return permissions

    @classmethod
    def _from_weaviate_role(cls, role: WeaviateRole) -> "Role":
        alias_permissions: List[AliasPermissionOutput] = []
        cluster_permissions: List[ClusterPermissionOutput] = []
        users_permissions: List[UsersPermissionOutput] = []
        collections_permissions: List[CollectionsPermissionOutput] = []
        roles_permissions: List[RolesPermissionOutput] = []
        data_permissions: List[DataPermissionOutput] = []
        backups_permissions: List[BackupsPermissionOutput] = []
        nodes_permissions: List[NodesPermissionOutput] = []
        tenants_permissions: List[TenantsPermissionOutput] = []
        replicate_permissions: List[ReplicatePermissionOutput] = []
        groups_permissions: List[GroupsPermissionOutput] = []

        for permission in role["permissions"]:
            if permission["action"] in ClusterAction.values():
                cluster_permissions.append(
                    ClusterPermissionOutput(actions={ClusterAction(permission["action"])})
                )
            elif permission["action"] in UsersAction.values():
                users = permission.get("users")
                if users is not None:
                    users_permissions.append(
                        UsersPermissionOutput(
                            actions={UsersAction(permission["action"])},
                            users=users["users"],
                        )
                    )
            elif permission["action"] in CollectionsAction.values():
                collections = permission.get("collections")
                if collections is not None:
                    collections_permissions.append(
                        CollectionsPermissionOutput(
                            collection=collections["collection"],
                            actions={CollectionsAction(permission["action"])},
                        )
                    )
            elif permission["action"] in TenantsAction.values():
                tenants = permission.get("tenants")
                if tenants is not None:
                    tenants_permissions.append(
                        TenantsPermissionOutput(
                            collection=tenants["collection"],
                            tenant=tenants.get("tenant", "*"),
                            actions={TenantsAction(permission["action"])},
                        )
                    )
            elif permission["action"] in RolesAction.values():
                roles = permission.get("roles")
                if roles is not None:
                    scope = roles.get("scope")
                    roles_permissions.append(
                        RolesPermissionOutput(
                            role=roles["role"],
                            actions={RolesAction(permission["action"])},
                            scope=RoleScope(scope) if scope else None,
                        )
                    )
            elif permission["action"] in DataAction.values():
                data = permission.get("data")
                if data is not None:
                    data_permissions.append(
                        DataPermissionOutput(
                            collection=data["collection"],
                            tenant=data.get("tenant", "*"),
                            actions={DataAction(permission["action"])},
                        )
                    )
            elif permission["action"] in BackupsAction.values():
                backups = permission.get("backups")
                if backups is not None:
                    backups_permissions.append(
                        BackupsPermissionOutput(
                            collection=backups["collection"],
                            actions={BackupsAction(permission["action"])},
                        )
                    )
            elif permission["action"] in NodesAction.values():
                nodes = permission.get("nodes")
                if nodes is not None:
                    nodes_permissions.append(
                        NodesPermissionOutput(
                            collection=nodes.get("collection", "*"),
                            verbosity=nodes["verbosity"],
                            actions={NodesAction(permission["action"])},
                        )
                    )
            elif permission["action"] in ReplicateAction.values():
                replicate = permission.get("replicate")
                if replicate is not None:
                    replicate_permissions.append(
                        ReplicatePermissionOutput(
                            collection=replicate["collection"],
                            shard=replicate.get("shard", "*"),
                            actions={ReplicateAction(permission["action"])},
                        )
                    )
            elif permission["action"] in AliasAction.values():
                aliases = permission.get("aliases")
                if aliases is not None:
                    alias_permissions.append(
                        AliasPermissionOutput(
                            alias=aliases["alias"],
                            collection=aliases["collection"],
                            actions={AliasAction(permission["action"])},
                        )
                    )
            elif permission["action"] in GroupAction.values():
                groups = permission.get("groups")
                if groups is not None:
                    groups_permissions.append(
                        GroupsPermissionOutput(
                            group=groups["group"],
                            group_type=groups["groupType"],
                            actions={GroupAction(permission["action"])},
                        )
                    )
            else:
                _Warnings.unknown_permission_encountered(permission)

        return cls(
            name=role["name"],
            alias_permissions=_join_permissions(alias_permissions),
            cluster_permissions=_join_permissions(cluster_permissions),
            users_permissions=_join_permissions(users_permissions),
            collections_permissions=_join_permissions(collections_permissions),
            roles_permissions=_join_permissions(roles_permissions),
            groups_permissions=_join_permissions(groups_permissions),
            data_permissions=_join_permissions(data_permissions),
            backups_permissions=_join_permissions(backups_permissions),
            nodes_permissions=_join_permissions(nodes_permissions),
            tenants_permissions=_join_permissions(tenants_permissions),
            replicate_permissions=_join_permissions(replicate_permissions),
        )


T = TypeVar("T", bound=_Permission)


def _join_permissions(permissions: List[T]) -> List[T]:
    # permissions with the same resource can be combined and then have multiple actions
    unified: Dict[str, int] = {}
    for i, perm in enumerate(permissions):
        resource = ""
        for field in perm.model_fields_set:
            if (
                field == "actions"
            ):  # action is the one field that is not part of the resource and which we want to combine
                continue
            resource += field + str(getattr(perm, field)) + "#"
        if resource in unified:
            permissions[unified[resource]].actions.add(perm.actions.pop())
        else:
            unified[resource] = i

    return_permission: List[T] = []
    for i in unified.values():
        return_permission.append(permissions[i])

    return return_permission


ActionsType = Union[_Action, Sequence[_Action]]


PermissionsInputType = Union[
    _Permission,
    Sequence[_Permission],
    Sequence[Sequence[_Permission]],
    Sequence[Union[_Permission, Sequence[_Permission]]],
]
PermissionsCreateType = List[_Permission]


class Actions:
    Alias = AliasAction
    Data = DataAction
    Collections = CollectionsAction
    Roles = RolesAction
    Cluster = ClusterAction
    Nodes = NodesAction
    Backups = BackupsAction
    Tenants = TenantsAction
    Users = UsersAction
    Replicate = ReplicateAction
    Groups = GroupAction


class GroupsPermissions:
    @staticmethod
    def oidc(
        *,
        group: Union[str, Sequence[str]],
        read: bool = False,
        assign_and_revoke: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(group, str):
            group = [group]
        for g in group:
            permission = _GroupsPermission(group=g, group_type="oidc", actions=set())

            if read:
                permission.actions.add(GroupAction.READ)
            if assign_and_revoke:
                permission.actions.add(GroupAction.ASSIGN_AND_REVOKE)

            if len(permission.actions) > 0:
                permissions.append(permission)

        return permissions


class NodesPermissions:
    @staticmethod
    def verbose(
        *,
        collection: Union[str, Sequence[str]],
        read: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        for c in collection:
            permission = _NodesPermission(collection=c, verbosity="verbose", actions=set())

            if read:
                permission.actions.add(NodesAction.READ)
            if len(permission.actions) > 0:
                permissions.append(permission)

        return permissions

    @staticmethod
    def minimal(
        *,
        read: bool = False,
    ) -> PermissionsCreateType:
        if read:
            permissions: List[_Permission] = [
                _NodesPermission(collection="*", verbosity="minimal", actions={NodesAction.READ})
            ]
            return permissions
        return []


class Permissions:
    Nodes = NodesPermissions
    Groups = GroupsPermissions

    @staticmethod
    def alias(
        *,
        alias: Union[str, Sequence[str]],
        collection: Union[str, Sequence[str]],
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(alias, str):
            alias = [alias]
        if isinstance(collection, str):
            collection = [collection]

        for a in alias:
            for c in collection:
                permission = _AliasPermission(alias=a, collection=c, actions=set())

                if create:
                    permission.actions.add(AliasAction.CREATE)
                if read:
                    permission.actions.add(AliasAction.READ)
                if update:
                    permission.actions.add(AliasAction.UPDATE)
                if delete:
                    permission.actions.add(AliasAction.DELETE)

                if len(permission.actions) > 0:
                    permissions.append(permission)
        return permissions

    @staticmethod
    def data(
        *,
        collection: Union[str, Sequence[str]],
        tenant: Union[str, Sequence[str], None] = None,
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        if tenant is None:
            tenant = ["*"]
        if isinstance(tenant, str):
            tenant = [tenant]
        for c in collection:
            for t in tenant:
                permission = _DataPermission(collection=c, tenant=t, actions=set())

                if create:
                    permission.actions.add(DataAction.CREATE)
                if read:
                    permission.actions.add(DataAction.READ)
                if update:
                    permission.actions.add(DataAction.UPDATE)
                if delete:
                    permission.actions.add(DataAction.DELETE)

                if len(permission.actions) > 0:
                    permissions.append(permission)
        return permissions

    @staticmethod
    def collections(
        *,
        collection: Union[str, Sequence[str]],
        create_collection: bool = False,
        read_config: bool = False,
        update_config: bool = False,
        delete_collection: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        for c in collection:
            permission = _CollectionsPermission(collection=c, actions=set())

            if create_collection:
                permission.actions.add(CollectionsAction.CREATE)
            if read_config:
                permission.actions.add(CollectionsAction.READ)
            if update_config:
                permission.actions.add(CollectionsAction.UPDATE)
            if delete_collection:
                permission.actions.add(CollectionsAction.DELETE)
            if len(permission.actions) > 0:
                permissions.append(permission)
        return permissions

    @staticmethod
    def tenants(
        *,
        collection: Union[str, Sequence[str]],
        tenant: Union[str, Sequence[str], None] = None,
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        if tenant is None:
            tenant = ["*"]
        if isinstance(tenant, str):
            tenant = [tenant]
        for c in collection:
            for t in tenant:
                permission = _TenantsPermission(collection=c, tenant=t, actions=set())

                if create:
                    permission.actions.add(TenantsAction.CREATE)
                if read:
                    permission.actions.add(TenantsAction.READ)
                if update:
                    permission.actions.add(TenantsAction.UPDATE)
                if delete:
                    permission.actions.add(TenantsAction.DELETE)

                if len(permission.actions) > 0:
                    permissions.append(permission)

        return permissions

    @staticmethod
    def replicate(
        *,
        collection: Union[str, Sequence[str]],
        shard: Union[str, Sequence[str], None] = None,
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        if shard is None:
            shard = ["*"]
        if isinstance(shard, str):
            shard = [shard]
        for c in collection:
            for s in shard:
                permission = _ReplicatePermission(collection=c, shard=s, actions=set())

                if create:
                    permission.actions.add(ReplicateAction.CREATE)
                if read:
                    permission.actions.add(ReplicateAction.READ)
                if update:
                    permission.actions.add(ReplicateAction.UPDATE)
                if delete:
                    permission.actions.add(ReplicateAction.DELETE)

                if len(permission.actions) > 0:
                    permissions.append(permission)

        return permissions

    @staticmethod
    def roles(
        *,
        role: Union[str, Sequence[str]],
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
        scope: Optional[RoleScope] = None,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(role, str):
            role = [role]
        for r in role:
            permission = _RolesPermission(role=r, actions=set())
            if read:
                permission.actions.add(RolesAction.READ)
            if create:
                permission.actions.add(RolesAction.CREATE)
            if update:
                permission.actions.add(RolesAction.UPDATE)
            if delete:
                permission.actions.add(RolesAction.DELETE)
            if scope is not None:
                permission.scope = scope.value
            if len(permission.actions) > 0:
                permissions.append(permission)

        return permissions

    @staticmethod
    def users(
        *,
        user: Union[str, Sequence[str]],
        create: bool = False,
        read: bool = False,
        update: bool = False,
        delete: bool = False,
        assign_and_revoke: bool = False,
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(user, str):
            user = [user]
        for u in user:
            permission = _UsersPermission(users=u, actions=set())

            if create:
                permission.actions.add(UsersAction.CREATE)
            if read:
                permission.actions.add(UsersAction.READ)
            if update:
                permission.actions.add(UsersAction.UPDATE)
            if delete:
                permission.actions.add(UsersAction.DELETE)
            if assign_and_revoke:
                permission.actions.add(UsersAction.ASSIGN_AND_REVOKE)

            if len(permission.actions) > 0:
                permissions.append(permission)

        return permissions

    @staticmethod
    def backup(
        *, collection: Union[str, Sequence[str]], manage: bool = False
    ) -> PermissionsCreateType:
        permissions: List[_Permission] = []
        if isinstance(collection, str):
            collection = [collection]
        for c in collection:
            permission = _BackupsPermission(collection=c, actions=set())

            if manage:
                permission.actions.add(BackupsAction.MANAGE)
            if len(permission.actions) > 0:
                permissions.append(permission)
        return permissions

    @staticmethod
    def cluster(*, read: bool = False) -> PermissionsCreateType:
        if read:
            return [_ClusterPermission(actions={ClusterAction.READ})]
        return []



import datetime
import sys
import uuid as uuid_package
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from typing_extensions import TypeAlias, deprecated

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_args, get_origin, get_type_hints
else:
    from typing import Annotated, get_args, get_origin, get_type_hints

from weaviate.collections.classes.generative import (
    _GenerativeConfigRuntime,
    _GenerativeConfigRuntimeOptions,
    _GroupedTask,
    _SinglePrompt,
    _to_text_array,
)
from weaviate.collections.classes.grpc import (
    METADATA,
    PROPERTIES,
    REFERENCES,
    GroupBy,
    MetadataQuery,
    QueryNested,
    Rerank,
    _QueryReference,
    _QueryReferenceMultiTarget,
)
from weaviate.collections.classes.types import (
    IReferences,
    M,
    P,
    Properties,
    R,
    References,
    TProperties,
    TReferences,
    WeaviateProperties,
    _WeaviateInput,
)
from weaviate.exceptions import (
    WeaviateInvalidInputError,
    WeaviateUnsupportedFeatureError,
)
from weaviate.proto.v1 import generative_pb2, search_get_pb2
from weaviate.types import INCLUDE_VECTOR, UUID, UUIDS
from weaviate.util import _ServerVersion, _to_beacons


@dataclass
class MetadataReturn:
    """Metadata of an object returned by a query."""

    creation_time: Optional[datetime.datetime] = None
    last_update_time: Optional[datetime.datetime] = None
    distance: Optional[float] = None
    certainty: Optional[float] = None
    score: Optional[float] = None
    explain_score: Optional[str] = None
    is_consistent: Optional[bool] = None
    rerank_score: Optional[float] = None

    def _is_empty(self) -> bool:
        return all(
            [
                self.creation_time is None,
                self.last_update_time is None,
                self.distance is None,
                self.certainty is None,
                self.score is None,
                self.explain_score is None,
                self.is_consistent is None,
                self.rerank_score is None,
            ]
        )


@dataclass
class GroupByMetadataReturn:
    """Metadata of an object returned by a group by query."""

    distance: Optional[float] = None


@dataclass
class _Object(Generic[P, R, M]):
    uuid: uuid_package.UUID
    metadata: M
    properties: P
    references: R
    vector: Dict[str, Union[List[float], List[List[float]]]]
    collection: str


@dataclass
class Object(Generic[P, R], _Object[P, R, MetadataReturn]):
    """A single Weaviate object returned by a query within the `.query` namespace of a collection."""


@dataclass
class MetadataSingleObjectReturn:
    """Metadata of an object returned by the `fetch_object_by_id` query."""

    creation_time: datetime.datetime
    last_update_time: datetime.datetime
    is_consistent: Optional[bool]


@dataclass
class ObjectSingleReturn(Generic[P, R], _Object[P, R, MetadataSingleObjectReturn]):
    """A single Weaviate object returned by the `fetch_object_by_id` query."""


@dataclass
class GroupByObject(Generic[P, R], _Object[P, R, GroupByMetadataReturn]):
    """A single Weaviate object returned by a query with the `group_by` argument specified."""

    belongs_to_group: str


GenerativeMetadata = Union[
    generative_pb2.GenerativeAnthropicMetadata,
    generative_pb2.GenerativeAnyscaleMetadata,
    generative_pb2.GenerativeAWSMetadata,
    generative_pb2.GenerativeCohereMetadata,
    generative_pb2.GenerativeDatabricksMetadata,
    generative_pb2.GenerativeDummyMetadata,
    generative_pb2.GenerativeFriendliAIMetadata,
    generative_pb2.GenerativeGoogleMetadata,
    generative_pb2.GenerativeMistralMetadata,
    generative_pb2.GenerativeNvidiaMetadata,
    generative_pb2.GenerativeOllamaMetadata,
    generative_pb2.GenerativeOpenAIMetadata,
]


@dataclass
class GenerativeSingle:
    """The generative data returned relevant to a single prompt generative query."""

    debug: Optional[generative_pb2.GenerativeDebug]
    metadata: Optional[GenerativeMetadata]
    text: Optional[str]


@dataclass
class GenerativeGrouped:
    """The generative data returned relevant to a grouped prompt generative query."""

    metadata: Optional[GenerativeMetadata]
    text: Optional[str]


class GenerativeObject(Generic[P, R], Object[P, R]):
    """A single Weaviate object returned by a query within the `generate` namespace of a collection."""

    __generated: Optional[str]
    generative: Optional[GenerativeSingle]

    # init required because of nuances of dataclass when defining @property generated and private var __generated
    def __init__(
        self,
        generated: Optional[str],
        generative: Optional[GenerativeSingle],
        uuid: uuid_package.UUID,
        metadata: MetadataReturn,
        properties: P,
        references: R,
        vector: Dict[str, Union[List[float], List[List[float]]]],
        collection: str,
    ) -> None:
        self.__generated = generated
        self.generative = generative
        super().__init__(
            uuid=uuid,
            metadata=metadata,
            properties=properties,
            references=references,
            vector=vector,
            collection=collection,
        )

    @property
    @deprecated(
        "The generated field is deprecated. Use generative.text instead.", category=None
    )  # todo: turn into a runtime warning in the future
    def generated(self) -> Optional[str]:
        """The single generated text of the object."""
        return self.__generated


class GenerativeReturn(Generic[P, R]):
    """The return type of a query within the `generate` namespace of a collection."""

    __generated: Optional[str]
    objects: List[GenerativeObject[P, R]]
    generative: Optional[GenerativeGrouped]

    # init required because of nuances of dataclass when defining @property generated and private var __generated
    def __init__(
        self,
        generated: Optional[str],
        objects: List[GenerativeObject[P, R]],
        generative: Optional[GenerativeGrouped],
    ) -> None:
        self.__generated = generated
        self.objects = objects
        self.generative = generative

    @property
    @deprecated(
        "The generated field is deprecated. Use generative.text instead.", category=None
    )  # todo: turn into a runtime warning in the future
    def generated(self) -> Optional[str]:
        """The grouped generated text of the objects."""
        return self.__generated


@dataclass
class Group(Generic[P, R]):
    """A group of objects returned in a group by query."""

    name: str
    min_distance: float
    max_distance: float
    number_of_objects: int
    objects: List[GroupByObject[P, R]]
    rerank_score: Optional[float]


@dataclass
class GenerativeGroup(Generic[P, R], Group[P, R]):
    """A group of objects returned in a generative group by query."""

    generated: Optional[str]


@dataclass
class GenerativeGroupByReturn(Generic[P, R]):
    """The return type of a query within the `.generate` namespace of a collection with the `group_by` argument specified."""

    objects: List[GroupByObject[P, R]]
    groups: Dict[str, GenerativeGroup[P, R]]
    generated: Optional[str]


@dataclass
class GroupByReturn(Generic[P, R]):
    """The return type of a query within the `.query` namespace of a collection with the `group_by` argument specified."""

    objects: List[GroupByObject[P, R]]
    groups: Dict[str, Group[P, R]]


@dataclass
class QueryReturn(Generic[P, R]):
    """The return type of a query within the `.query` namespace of a collection."""

    objects: List[Object[P, R]]


_GQLEntryReturnType: TypeAlias = Dict[str, List[Dict[str, Any]]]


@dataclass
class _RawGQLReturn:
    aggregate: _GQLEntryReturnType
    explore: _GQLEntryReturnType
    get: _GQLEntryReturnType
    errors: Optional[Dict[str, Any]]


class _Generative:
    single: Union[str, _SinglePrompt, None]
    grouped: Union[str, _GroupedTask, None]
    grouped_properties: Optional[List[str]]
    generative_provider: Optional[_GenerativeConfigRuntime]

    def __init__(
        self,
        single: Union[str, _SinglePrompt, None],
        grouped: Union[str, _GroupedTask, None],
        grouped_properties: Optional[List[str]],
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
    ) -> None:
        self.single = single
        self.grouped = grouped
        self.grouped_properties = grouped_properties
        self.generative_provider = generative_provider

    def to_grpc(self, server_version: _ServerVersion) -> generative_pb2.GenerativeSearch:
        if server_version.is_lower_than(1, 27, 14):
            if self.generative_provider is not None:
                raise WeaviateUnsupportedFeatureError("Dynamic RAG", str(server_version), "1.27.14")

            if isinstance(self.single, _SinglePrompt):
                single_prompt: Optional[str] = self.single.prompt
            else:
                single_prompt = self.single

            if isinstance(self.grouped, _GroupedTask):
                grouped_task: Optional[str] = self.grouped.prompt
                grouped_properties = self.grouped.non_blob_properties
            else:
                grouped_task = self.grouped
                grouped_properties = self.grouped_properties

            return generative_pb2.GenerativeSearch(
                single_response_prompt=single_prompt,
                grouped_response_task=grouped_task,
                grouped_properties=grouped_properties,
            )
        else:
            single: Optional[generative_pb2.GenerativeSearch.Single] = None
            if isinstance(self.single, _SinglePrompt):
                single = generative_pb2.GenerativeSearch.Single(
                    prompt=self.single.prompt,
                    queries=(
                        [
                            self.generative_provider._to_grpc(
                                _GenerativeConfigRuntimeOptions(
                                    self.single.metadata,
                                    self.single.images,
                                    self.single.image_properties,
                                )
                            )
                        ]
                        if self.generative_provider is not None
                        else None
                    ),
                    debug=self.single.debug,
                )
            if isinstance(self.single, str):
                single = generative_pb2.GenerativeSearch.Single(
                    prompt=self.single,
                    queries=(
                        [self.generative_provider._to_grpc(_GenerativeConfigRuntimeOptions())]
                        if self.generative_provider is not None
                        else None
                    ),
                )

            grouped: Optional[generative_pb2.GenerativeSearch.Grouped] = None
            if isinstance(self.grouped, _GroupedTask):
                grouped = generative_pb2.GenerativeSearch.Grouped(
                    task=self.grouped.prompt,
                    properties=_to_text_array(self.grouped.non_blob_properties),
                    queries=(
                        [
                            self.generative_provider._to_grpc(
                                _GenerativeConfigRuntimeOptions(
                                    self.grouped.metadata,
                                    self.grouped.images,
                                    self.grouped.image_properties,
                                )
                            )
                        ]
                        if self.generative_provider is not None
                        else None
                    ),
                )
            if isinstance(self.grouped, str):
                grouped = generative_pb2.GenerativeSearch.Grouped(
                    task=self.grouped,
                    properties=(
                        _to_text_array(self.grouped_properties)
                        if self.grouped_properties is not None
                        else None
                    ),
                    queries=(
                        [self.generative_provider._to_grpc(_GenerativeConfigRuntimeOptions())]
                        if self.generative_provider is not None
                        else None
                    ),
                )

            return generative_pb2.GenerativeSearch(
                single=single,
                grouped=grouped,
            )


class _GroupBy:
    prop: str
    number_of_groups: int
    objects_per_group: int

    def __init__(self, prop: str, number_of_groups: int, objects_per_group: int) -> None:
        self.prop = prop
        self.number_of_groups = number_of_groups
        self.objects_per_group = objects_per_group

    def to_grpc(self) -> search_get_pb2.GroupBy:
        return search_get_pb2.GroupBy(
            path=[self.prop],
            number_of_groups=self.number_of_groups,
            objects_per_group=self.objects_per_group,
        )

    @classmethod
    def from_input(cls, group_by: Optional[GroupBy]) -> Optional["_GroupBy"]:
        return (
            cls(
                prop=group_by.prop,
                number_of_groups=group_by.number_of_groups,
                objects_per_group=group_by.objects_per_group,
            )
            if group_by
            else None
        )


Nested = Annotated[P, "NESTED"]


def __is_nested(value: Any) -> bool:
    return (
        get_origin(value) is Annotated
        and len(get_args(value)) == 2
        and cast(str, get_args(value)[1]) == "NESTED"
    )


def __create_nested_property_from_nested(name: str, value: Any) -> QueryNested:
    inner_type = get_args(value)[0]
    # If this nested property contains an object array, use the element type
    if get_origin(inner_type) is list:
        inner_type = get_args(inner_type)[0]

    return QueryNested(
        name=name,
        properties=[
            __create_nested_property_from_nested(key, val) if __is_nested(val) else key
            for key, val in get_type_hints(inner_type, include_extras=True).items()
        ],
    )


class _Reference:
    def __init__(
        self,
        target_collection: Optional[str],
        uuids: UUIDS,
    ):
        """You should not initialise this class directly. Use the `.to_multi()` class methods instead."""
        self.__target_collection = target_collection if target_collection else ""
        self.__uuids = uuids

    def _to_beacons(self) -> List[Dict[str, str]]:
        return _to_beacons(self.__uuids, self.__target_collection)

    @property
    def is_one_to_many(self) -> bool:
        """Returns True if the reference is to a one-to-many references, i.e. points to more than one object."""
        return self.__uuids is not None and isinstance(self.__uuids, list) and len(self.__uuids) > 1


class ReferenceToMulti(_WeaviateInput):
    """Use this class when you want to insert a multi-target reference property."""

    target_collection: str
    uuids: UUIDS

    def _to_beacons(self) -> List[Dict[str, str]]:
        return _to_beacons(self.uuids, self.target_collection)

    @property
    def uuids_str(self) -> List[str]:
        """Returns the UUIDs as strings."""
        if isinstance(self.uuids, list):
            return [str(uid) for uid in self.uuids]
        else:
            return [str(self.uuids)]


class _CrossReference(Generic[Properties, IReferences]):
    def __init__(
        self,
        objects: Optional[List[Object[Properties, IReferences]]],
    ):
        self.__objects = objects

    @classmethod
    def _from(
        cls, objects: List[Object[Properties, IReferences]]
    ) -> "_CrossReference[Properties, IReferences]":
        return cls(objects)

    @property
    def objects(self) -> List[Object[Properties, IReferences]]:
        """Returns the objects of the cross reference."""
        return self.__objects or []


CrossReference: TypeAlias = _CrossReference[Properties, IReferences]
"""Use this TypeAlias when you want to type hint a cross reference within a generic data model.

If you want to define a reference property when creating your collection, use `ReferenceProperty` or `ReferencePropertyMultiTarget` instead.

If you want to create a reference when inserting an object, supply the UUIDs directly or use `Reference.to_multi()` instead.

Example:
    >>> import typing
    >>> import weaviate.classes as wvc
    >>>
    >>> class One(typing.TypedDict):
    ...     prop: str
    >>>
    >>> class Two(typing.TypedDict):
    ...     one: wvc.CrossReference[One]
"""

CrossReferences = Mapping[str, _CrossReference[WeaviateProperties, "CrossReferences"]]


SingleReferenceInput = Union[UUID, ReferenceToMulti]

ReferenceInput: TypeAlias = Union[UUID, Sequence[UUID], ReferenceToMulti]
"""This type alias is used when providing references as inputs within the `.data` namespace of a collection."""
ReferenceInputs: TypeAlias = Mapping[str, ReferenceInput]
"""This type alias is used when providing references as inputs within the `.data` namespace of a collection."""


@dataclass
class CrossReferenceAnnotation:
    """Dataclass to be used when annotating a generic cross reference property with options for retrieving data from the cross referenced object when querying.

    Example:
        >>> import typing
        >>> import weaviate.classes as wvc
        >>>
        >>> class One(typing.TypedDict):
        ...     prop: str
        >>>
        >>> class Two(typing.TypedDict):
        ...     one: typing.Annotated[
        ...         wvc.CrossReference[One],
        ...         wvc.CrossReferenceAnnotation(include_vector=True)
        ...     ]
    """

    include_vector: bool = field(default=False)
    metadata: Optional[MetadataQuery] = field(default=None)
    target_collection: Optional[str] = field(default=None)


def _extract_types_from_reference(
    type_: CrossReference[Properties, "References"], field: str
) -> Tuple[Type[Properties], Type["References"]]:
    """Extract first inner type from CrossReference[Properties, References]."""
    if get_origin(type_) == _CrossReference:
        return cast(Tuple[Type[Properties], Type[References]], get_args(type_))
    raise WeaviateInvalidInputError(
        f"Type: {type_} of field: {field} is not CrossReference[Properties, References]"
    )


def _extract_types_from_annotated_reference(
    type_: Annotated[CrossReference[Properties, "References"], CrossReferenceAnnotation],
    field: str,
) -> Tuple[Type[Properties], Type["References"]]:
    """Extract inner type from Annotated[CrossReference[Properties, References]]."""
    assert get_origin(type_) is Annotated, f"field: {field} with type: {type_} must be annotated"
    args = get_args(type_)
    inner_type = cast(CrossReference[Properties, References], args[0])
    return _extract_types_from_reference(inner_type, field)


def __is_annotated_reference(value: Any) -> bool:
    return (
        get_origin(value) is Annotated
        and len(get_args(value)) == 2
        and get_origin(get_args(value)[0]) is _CrossReference
    )


def __create_link_to_from_annotated_reference(
    link_on: str,
    value: Annotated[CrossReference[Properties, "References"], CrossReferenceAnnotation],
) -> Union[_QueryReference, _QueryReferenceMultiTarget]:
    """Create FromReference or FromReferenceMultiTarget from Annotated[CrossReference[Properties], ReferenceAnnotation]."""
    assert get_origin(value) is Annotated, (
        f"field: {link_on} with type: {value} must be Annotated[CrossReference]"
    )
    args = cast(List[CrossReference[Properties, References]], get_args(value))
    inner_type = args[0]
    assert get_origin(inner_type) is _CrossReference, (
        f"field: {link_on} with inner_type: {inner_type} must be CrossReference"
    )
    inner_type_metadata = cast(
        Tuple[CrossReferenceAnnotation], getattr(value, "__metadata__", None)
    )
    annotation = inner_type_metadata[0]
    types = _extract_types_from_annotated_reference(value, link_on)
    if annotation.target_collection is not None:
        return _QueryReferenceMultiTarget(
            link_on=link_on,
            include_vector=annotation.include_vector,
            return_metadata=annotation.metadata,
            return_properties=_extract_properties_from_data_model(types[0]),
            return_references=_extract_references_from_data_model(types[1]),
            target_collection=annotation.target_collection,
        )
    else:
        return _QueryReference(
            link_on=link_on,
            include_vector=annotation.include_vector,
            return_metadata=annotation.metadata,
            return_properties=_extract_properties_from_data_model(types[0]),
            return_references=_extract_references_from_data_model(types[1]),
        )


def __create_link_to_from_reference(
    link_on: str,
    value: CrossReference[Properties, "References"],
) -> _QueryReference:
    """Create FromReference from CrossReference[Properties]."""
    types = _extract_types_from_reference(value, link_on)
    return _QueryReference(
        link_on=link_on,
        return_properties=_extract_properties_from_data_model(types[0]),
        return_references=_extract_references_from_data_model(types[1]),
    )


def _extract_properties_from_data_model(type_: Type[Properties]) -> PROPERTIES:
    """Extract properties of Properties recursively from Properties.

    Checks to see if there is a _Reference[Properties], Annotated[_Reference[Properties]], or _Nested[Properties]
    in the data model and lists out the properties as classes readily consumable by the underlying API.
    """
    return [
        __create_nested_property_from_nested(key, value) if __is_nested(value) else key
        for key, value in get_type_hints(type_, include_extras=True).items()
    ]


def _extract_references_from_data_model(
    type_: Type["References"],
) -> Optional[REFERENCES]:
    """Extract references of References recursively from References.

    Checks to see if there is a _Reference[References], Annotated[_Reference[References]], or _Nested[References]
    in the data model and lists out the references as classes readily consumable by the underlying API.
    """
    refs = [
        (
            __create_link_to_from_annotated_reference(key, value)
            if __is_annotated_reference(value)
            else __create_link_to_from_reference(key, value)
        )
        for key, value in get_type_hints(type_, include_extras=True).items()
    ]
    return refs if len(refs) > 0 else None


ReturnProperties: TypeAlias = Union[PROPERTIES, bool, Type[TProperties]]
ReturnReferences: TypeAlias = Union[
    Union[_QueryReference, Sequence[_QueryReference]], Type[TReferences]
]


@dataclass
class _QueryOptions:
    include_metadata: bool
    include_properties: bool
    include_references: bool
    include_vector: bool
    is_group_by: bool

    @classmethod
    def from_input(
        cls,
        return_metadata: Optional[METADATA],
        return_properties: Optional[ReturnProperties[Any]],
        include_vector: INCLUDE_VECTOR,
        collection_references: Optional[Type[Any]],
        query_references: Optional[ReturnReferences[Any]],
        rerank: Optional[Rerank] = None,
        group_by: Optional[GroupBy] = None,
    ) -> "_QueryOptions":
        return cls(
            include_metadata=return_metadata is not None or rerank is not None,
            include_properties=not (
                isinstance(return_properties, list) and len(return_properties) == 0
            ),
            include_references=collection_references is not None or query_references is not None,
            include_vector=include_vector if isinstance(include_vector, bool) else True,
            is_group_by=group_by is not None,
        )


QuerySingleReturn = Union[
    ObjectSingleReturn[Properties, References],
    ObjectSingleReturn[TProperties, TReferences],
    ObjectSingleReturn[Properties, CrossReferences],
    ObjectSingleReturn[Properties, TReferences],
    ObjectSingleReturn[TProperties, References],
    ObjectSingleReturn[TProperties, CrossReferences],
    None,
]

GenerativeGroupByReturnType = Union[
    GenerativeGroupByReturn[Properties, References],
    GenerativeGroupByReturn[TProperties, TReferences],
    GenerativeGroupByReturn[Properties, CrossReferences],
    GenerativeGroupByReturn[Properties, TReferences],
    GenerativeGroupByReturn[TProperties, References],
    GenerativeGroupByReturn[TProperties, CrossReferences],
]

GenerativeReturnType = Union[
    GenerativeReturn[Properties, References],
    GenerativeReturn[TProperties, TReferences],
    GenerativeReturn[Properties, CrossReferences],
    GenerativeReturn[Properties, TReferences],
    GenerativeReturn[TProperties, References],
    GenerativeReturn[TProperties, CrossReferences],
]

# The way in which generic type aliases work requires that all the generic arguments
# are listed first and in the order of their appearance in the typealias.
# GenerativeNearMediaReturn[Properties, References, TProperties, TReferences] is the intended use and so
# these four generics appear first. All others resolve afterwards correctly
GenerativeNearMediaReturnType = Union[
    GenerativeReturnType[Properties, References, TProperties, TReferences],
    GenerativeGroupByReturnType[Properties, References, TProperties, TReferences],
]
"""@Deprecated: Use `GenerativeSearchReturnType` instead."""

GenerativeSearchReturnType = Union[
    GenerativeReturnType[Properties, References, TProperties, TReferences],
    GenerativeGroupByReturnType[Properties, References, TProperties, TReferences],
]

QueryReturnType = Union[
    QueryReturn[Properties, References],
    QueryReturn[TProperties, TReferences],
    QueryReturn[Properties, CrossReferences],
    QueryReturn[Properties, TReferences],
    QueryReturn[TProperties, References],
    QueryReturn[TProperties, CrossReferences],
]

GroupByReturnType = Union[
    GroupByReturn[Properties, References],
    GroupByReturn[TProperties, TReferences],
    GroupByReturn[Properties, CrossReferences],
    GroupByReturn[Properties, TReferences],
    GroupByReturn[TProperties, References],
    GroupByReturn[TProperties, CrossReferences],
]

QuerySearchReturnType = Union[
    QueryReturnType[Properties, References, TProperties, TReferences],
    GroupByReturnType[Properties, References, TProperties, TReferences],
]

QueryNearMediaReturnType = Union[
    QueryReturnType[Properties, References, TProperties, TReferences],
    GroupByReturnType[Properties, References, TProperties, TReferences],
]
"""@Deprecated: Use `QuerySearchReturnType` instead."""



import uuid as uuid_lib
from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TypeAlias

from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import (
    PROPERTIES,
    PROPERTY,
    REFERENCE,
    REFERENCES,
    BM25OperatorOptions,
    BM25OperatorOr,
    HybridFusion,
    HybridVectorType,
    Move,
    NearVectorInputType,
    QueryNested,
    Rerank,
    TargetVectorJoinType,
    _MetadataQuery,
    _QueryReference,
    _QueryReferenceMultiTarget,
    _Sorting,
)
from weaviate.collections.classes.internal import (
    _Generative,
    _GroupBy,
)
from weaviate.collections.filters import _FilterToGRPC
from weaviate.collections.grpc.shared import _BaseGRPC
from weaviate.proto.v1 import base_search_pb2, search_get_pb2
from weaviate.types import NUMBER, UUID
from weaviate.util import _ServerVersion
from weaviate.validator import _validate_input, _ValidateArgument

# Can be found in the google.protobuf.internal.well_known_types.pyi stub file but is defined explicitly here for clarity.
_PyValue: TypeAlias = Union[
    Dict[str, "_PyValue"],
    List["_PyValue"],
    str,
    float,
    bool,
    None,
    List[float],
    List[int],
    List[str],
    List[bool],
    List[UUID],
]


@dataclass
class _Move:
    force: float
    concepts: List[str]
    objects: List[uuid_lib.UUID]


A = TypeVar("A")


class _QueryGRPC(_BaseGRPC):
    def __init__(
        self,
        weaviate_version: _ServerVersion,
        name: str,
        tenant: Optional[str],
        consistency_level: Optional[ConsistencyLevel],
        validate_arguments: bool,
        uses_125_api: bool,
        uses_127_api: bool,
    ):
        super().__init__(weaviate_version, consistency_level, validate_arguments)
        self._name: str = name
        self._tenant = tenant
        self._validate_arguments = validate_arguments
        self.__uses_125_api = uses_125_api
        self.__uses_127_api = uses_127_api

    def __parse_near_options(
        self,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
    ) -> Tuple[Optional[float], Optional[float]]:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([float, int, None], "certainty", certainty),
                    _ValidateArgument([float, int, None], "distance", distance),
                ]
            )
        return (
            float(certainty) if certainty is not None else None,
            float(distance) if distance is not None else None,
        )

    def get(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[_Sorting] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
    ) -> search_get_pb2.SearchRequest:
        if self._validate_arguments:
            _validate_input(_ValidateArgument([_Sorting, None], "sort", sort))

        if sort is not None:
            sort_by: Optional[List[search_get_pb2.SortBy]] = [
                search_get_pb2.SortBy(ascending=sort.ascending, path=[sort.prop])
                for sort in sort.sorts
            ]
        else:
            sort_by = None

        return self.__create_request(
            after=after,
            limit=limit,
            offset=offset,
            filters=filters,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            sort_by=sort_by,
        )

    def hybrid(
        self,
        *,
        query: Optional[str],
        alpha: Optional[float] = None,
        vector: Optional[HybridVectorType] = None,
        properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        bm25_operator: Optional[BM25OperatorOptions] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
    ) -> search_get_pb2.SearchRequest:
        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            group_by=group_by,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            hybrid_search=self._parse_hybrid(
                query,
                alpha,
                vector,
                properties,
                bm25_operator,
                fusion_type,
                distance,
                target_vector,
            ),
        )

    def bm25(
        self,
        *,
        query: Optional[str],
        properties: Optional[List[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        operator: Optional[BM25OperatorOptions] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
    ) -> search_get_pb2.SearchRequest:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([None, str], "query", query),
                    _ValidateArgument([List, None], "properties", properties),
                ]
            )

        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            group_by=group_by,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            bm25=(
                base_search_pb2.BM25(
                    query=query,
                    properties=properties if properties is not None else [],
                    search_operator=base_search_pb2.SearchOperatorOptions(
                        operator=operator.operator,
                        minimum_or_tokens_match=operator.minimum_should_match
                        if isinstance(operator, BM25OperatorOr)
                        else None,
                    )
                    if operator is not None
                    else None,
                )
                if query is not None
                else None
            ),
        )

    def near_vector(
        self,
        *,
        near_vector: NearVectorInputType,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
    ) -> search_get_pb2.SearchRequest:
        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            group_by=group_by,
            near_vector=self._parse_near_vector(
                near_vector, certainty, distance, target_vector=target_vector
            ),
        )

    def near_object(
        self,
        *,
        near_object: UUID,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
    ) -> search_get_pb2.SearchRequest:
        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            group_by=group_by,
            near_object=self._parse_near_object(near_object, certainty, distance, target_vector),
        )

    def near_text(
        self,
        *,
        near_text: Union[List[str], str],
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        move_to: Optional[Move] = None,
        move_away: Optional[Move] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
    ) -> search_get_pb2.SearchRequest:
        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            group_by=group_by,
            near_text=self._parse_near_text(
                near_text,
                certainty,
                distance,
                move_away=move_away,
                move_to=move_to,
                target_vector=target_vector,
            ),
        )

    def near_media(
        self,
        *,
        media: str,
        type_: Literal["audio", "depth", "image", "imu", "thermal", "video"],
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        return_metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
    ) -> search_get_pb2.SearchRequest:
        return self.__create_request(
            limit=limit,
            offset=offset,
            filters=filters,
            metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
            generative=generative,
            rerank=rerank,
            autocut=autocut,
            group_by=group_by,
            **self._parse_media(
                media,
                type_,
                certainty,
                distance,
                target_vector,
            ),
        )

    def __create_request(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        metadata: Optional[_MetadataQuery] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Optional[REFERENCES] = None,
        generative: Optional[_Generative] = None,
        rerank: Optional[Rerank] = None,
        autocut: Optional[int] = None,
        group_by: Optional[_GroupBy] = None,
        near_vector: Optional[base_search_pb2.NearVector] = None,
        sort_by: Optional[Sequence[search_get_pb2.SortBy]] = None,
        hybrid_search: Optional[base_search_pb2.Hybrid] = None,
        bm25: Optional[base_search_pb2.BM25] = None,
        near_object: Optional[base_search_pb2.NearObject] = None,
        near_text: Optional[base_search_pb2.NearTextSearch] = None,
        near_audio: Optional[base_search_pb2.NearAudioSearch] = None,
        near_depth: Optional[base_search_pb2.NearDepthSearch] = None,
        near_image: Optional[base_search_pb2.NearImageSearch] = None,
        near_imu: Optional[base_search_pb2.NearIMUSearch] = None,
        near_thermal: Optional[base_search_pb2.NearThermalSearch] = None,
        near_video: Optional[base_search_pb2.NearVideoSearch] = None,
    ) -> search_get_pb2.SearchRequest:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([int, None], "limit", limit),
                    _ValidateArgument([int, None], "offset", offset),
                    _ValidateArgument([uuid_lib.UUID, str, None], "after", after),
                    _ValidateArgument([_Filters, None], "filters", filters),
                    _ValidateArgument([_MetadataQuery, None], "metadata", metadata),
                    _ValidateArgument([_Generative, None], "generative", generative),
                    _ValidateArgument([Rerank, None], "rerank", rerank),
                    _ValidateArgument([int, None], "autocut", autocut),
                    _ValidateArgument([_GroupBy, None], "group_by", group_by),
                    _ValidateArgument(
                        [str, bool, QueryNested, Sequence, None],
                        "return_properties",
                        return_properties,
                    ),
                    _ValidateArgument(
                        [_QueryReference, Sequence, None],
                        "return_references",
                        return_references,
                    ),
                ]
            )
            if isinstance(return_properties, Sequence):
                for prop in return_properties:
                    _validate_input(
                        _ValidateArgument(
                            expected=[str, QueryNested],
                            name="return_properties",
                            value=prop,
                        )
                    )

            if isinstance(return_references, Sequence):
                for ref in return_references:
                    _validate_input(
                        _ValidateArgument(
                            expected=[_QueryReference],
                            name="return_references",
                            value=ref,
                        )
                    )

        if return_references is not None:
            return_references_parsed: Optional[Set[REFERENCE]] = self.__convert_to_set(
                return_references
            )
        else:
            return_references_parsed = None

        return_properties_parsed = self.__parse_return_properties(return_properties)

        return search_get_pb2.SearchRequest(
            uses_123_api=True,
            uses_125_api=self.__uses_125_api,
            uses_127_api=self.__uses_127_api,
            collection=self._name,
            limit=limit,
            offset=offset,
            after=str(after) if after is not None else "",
            autocut=autocut,
            properties=self._translate_properties_from_python_to_grpc(
                return_properties_parsed, return_references_parsed
            ),
            metadata=(self._metadata_to_grpc(metadata) if metadata is not None else None),
            consistency_level=self._consistency_level,
            tenant=self._tenant,
            filters=_FilterToGRPC.convert(filters),
            generative=(
                generative.to_grpc(self._weaviate_version) if generative is not None else None
            ),
            group_by=group_by.to_grpc() if group_by is not None else None,
            rerank=(
                search_get_pb2.Rerank(property=rerank.prop, query=rerank.query)
                if rerank is not None
                else None
            ),
            near_vector=near_vector,
            sort_by=sort_by,
            hybrid_search=hybrid_search,
            bm25_search=bm25,
            near_object=near_object,
            near_text=near_text,
            near_audio=near_audio,
            near_depth=near_depth,
            near_image=near_image,
            near_imu=near_imu,
            near_thermal=near_thermal,
            near_video=near_video,
        )

    def _metadata_to_grpc(self, metadata: _MetadataQuery) -> search_get_pb2.MetadataRequest:
        return search_get_pb2.MetadataRequest(
            uuid=metadata.uuid,
            vector=metadata.vector,
            creation_time_unix=metadata.creation_time_unix,
            last_update_time_unix=metadata.last_update_time_unix,
            distance=metadata.distance,
            certainty=metadata.certainty,
            explain_score=metadata.explain_score,
            score=metadata.score,
            is_consistent=metadata.is_consistent,
            vectors=metadata.vectors,
        )

    def __resolve_property(self, prop: QueryNested) -> search_get_pb2.ObjectPropertiesRequest:
        props = prop.properties if isinstance(prop.properties, list) else [prop.properties]
        return search_get_pb2.ObjectPropertiesRequest(
            prop_name=prop.name,
            primitive_properties=[p for p in props if isinstance(p, str)],
            object_properties=[
                self.__resolve_property(p) for p in props if isinstance(p, QueryNested)
            ],
        )

    def __parse_return_properties(
        self, props: Union[PROPERTIES, bool, None]
    ) -> Optional[Set[PROPERTY]]:
        if props is None or props is True:
            return None
        return self.__convert_to_set([] if props is False else props)

    def _translate_properties_from_python_to_grpc(
        self, properties: Optional[Set[PROPERTY]], references: Optional[Set[REFERENCE]]
    ) -> Optional[search_get_pb2.PropertiesRequest]:
        if properties is None and references is None:
            return None
        return search_get_pb2.PropertiesRequest(
            return_all_nonref_properties=properties is None,
            non_ref_properties=(
                None
                if properties is None
                else [prop for prop in properties if isinstance(prop, str)]
            ),
            ref_properties=(
                None
                if references is None
                else [
                    search_get_pb2.RefPropertiesRequest(
                        reference_property=ref.link_on,
                        properties=self._translate_properties_from_python_to_grpc(
                            self.__parse_return_properties(ref.return_properties),
                            (
                                None
                                if ref.return_references is None
                                else self.__convert_to_set(ref.return_references)
                            ),
                        ),
                        metadata=(
                            self._metadata_to_grpc(ref._return_metadata)
                            if ref._return_metadata is not None
                            else None
                        ),
                        target_collection=(
                            ref.target_collection
                            if isinstance(ref, _QueryReferenceMultiTarget)
                            else None
                        ),
                    )
                    for ref in references
                ]
            ),
            object_properties=(
                None
                if properties is None
                else [
                    self.__resolve_property(prop)
                    for prop in properties
                    if isinstance(prop, QueryNested)
                ]
            ),
        )

    @staticmethod
    def __convert_to_set(args: Union[A, Sequence[A]]) -> Set[A]:
        if isinstance(args, list):
            return set(args)
        else:
            return {cast(A, args)}



# make sure to import all classes that should be available in the weaviate module
from . import (
    aggregate,
    backup,
    batch,
    config,
    data,
    generate,
    generics,
    init,
    query,
    rbac,
    replication,
    tenants,
)  # noqa: F401
from .config import ConsistencyLevel

__all__ = [
    "aggregate",
    "backup",
    "batch",
    "config",
    "ConsistencyLevel",
    "data",
    "generate",
    "generics",
    "init",
    "query",
    "tenants",
    "rbac",
    "replication",
]



"""
Client class definition.
"""

from typing import Any, Dict, Optional, Tuple, Union

from weaviate.aliases.async_ import _AliasAsync
from weaviate.aliases.sync import _Alias
from weaviate.client_executor import _WeaviateClientExecutor
from weaviate.collections.classes.internal import _RawGQLReturn
from weaviate.collections.collections.async_ import _CollectionsAsync
from weaviate.collections.collections.sync import _Collections
from weaviate.connect.v4 import ConnectionAsync, ConnectionSync
from weaviate.groups.async_ import _GroupsAsync
from weaviate.groups.sync import _Groups
from weaviate.users.async_ import _UsersAsync
from weaviate.users.sync import _Users

from .backup import _Backup, _BackupAsync
from .cluster import _Cluster, _ClusterAsync
from .collections.batch.client import _BatchClientWrapper
from .debug import _Debug, _DebugAsync
from .rbac import _Roles, _RolesAsync
from .types import NUMBER

TIMEOUT_TYPE = Union[Tuple[NUMBER, NUMBER], NUMBER]

class WeaviateAsyncClient(_WeaviateClientExecutor[ConnectionAsync]):
    _connection: ConnectionAsync
    alias: _AliasAsync
    backup: _BackupAsync
    collections: _CollectionsAsync
    cluster: _ClusterAsync
    debug: _DebugAsync
    groups: _GroupsAsync
    roles: _RolesAsync
    users: _UsersAsync

    async def close(self) -> None: ...
    async def connect(self) -> None: ...
    def is_connected(self) -> bool: ...
    async def is_live(self) -> bool: ...
    async def is_ready(self) -> bool: ...
    async def graphql_raw_query(self, gql_query: str) -> _RawGQLReturn: ...
    async def get_meta(self) -> dict: ...
    async def get_open_id_configuration(self) -> Optional[Dict[str, Any]]: ...
    async def __aenter__(self) -> "WeaviateAsyncClient": ...
    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...

class WeaviateClient(_WeaviateClientExecutor[ConnectionSync]):
    _connection: ConnectionSync
    alias: _Alias
    backup: _Backup
    batch: _BatchClientWrapper
    collections: _Collections
    cluster: _Cluster
    debug: _Debug
    groups: _Groups
    roles: _Roles
    users: _Users

    def close(self) -> None: ...
    def connect(self) -> None: ...
    def is_connected(self) -> bool: ...
    def is_live(self) -> bool: ...
    def is_ready(self) -> bool: ...
    def graphql_raw_query(self, gql_query: str) -> _RawGQLReturn: ...
    def get_meta(self) -> dict: ...
    def get_open_id_configuration(self) -> Optional[Dict[str, Any]]: ...
    def __enter__(self) -> "WeaviateClient": ...
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...

class Client:
    def __init__(self) -> None: ...



from dataclasses import dataclass
from typing import Dict, Final, List, Literal

from weaviate.rbac.models import (
    Role,
    UserTypes,
)

USER_TYPE_DB: Final = "db"
USER_TYPE_OIDC: Final = "oidc"
USER_TYPE = Literal["db", "oidc"]


@dataclass
class OwnUser:
    user_id: str
    roles: Dict[str, Role]
    groups: List[str]


@dataclass
class UserBase:
    user_id: str
    role_names: List[str]
    user_type: UserTypes


@dataclass
class UserDB(UserBase):
    user_type: UserTypes
    active: bool


@dataclass
class UserOIDC(UserBase):
    user_type: UserTypes = UserTypes.OIDC



from __future__ import annotations

from typing import Awaitable, Callable, Dict, List, Optional, Union

import httpx
from authlib.integrations.httpx_client import (  # type: ignore
    AsyncOAuth2Client,
    OAuth2Client,
)

from weaviate.auth import (
    AuthBearerToken,
    AuthClientCredentials,
    AuthClientPassword,
    AuthCredentials,
)
from weaviate.exceptions import AuthenticationFailedError, MissingScopeError

from ..util import _decode_json_response_dict
from ..warnings import _Warnings
from . import executor

AUTH_DEFAULT_TIMEOUT = 5
OIDC_CONFIG = Dict[str, Union[str, List[str]]]

Result = Union[OAuth2Client, Awaitable[AsyncOAuth2Client]]
MountsMaker = Union[
    Callable[[], Dict[str, httpx.AsyncHTTPTransport]],
    Callable[[], Dict[str, httpx.HTTPTransport]],
]


class _Auth:
    def __init__(
        self,
        oidc_config: OIDC_CONFIG,
        credentials: AuthCredentials,
        make_mounts: MountsMaker,
        colour: executor.Colour,
    ) -> None:
        self._credentials: AuthCredentials = credentials
        self.__make_mounts = make_mounts
        self.__colour: executor.Colour = colour
        config_url = oidc_config["href"]
        client_id = oidc_config["clientId"]
        assert isinstance(config_url, str) and isinstance(client_id, str)
        self._open_id_config_url: str = config_url
        self._client_id: str = client_id
        self._default_scopes: List[str] = []
        if "scopes" in oidc_config:
            default_scopes = oidc_config["scopes"]
            assert isinstance(default_scopes, list)
            self._default_scopes = default_scopes

        self._token_endpoint: Optional[str] = None
        self._oidc_config = oidc_config

    @staticmethod
    def result(result: Result) -> OAuth2Client:
        assert isinstance(result, OAuth2Client)
        return result

    @staticmethod
    async def aresult(result: Result) -> AsyncOAuth2Client:
        assert isinstance(result, Awaitable)
        return await result

    @staticmethod
    def use(
        oidc_config: OIDC_CONFIG,
        credentials: AuthCredentials,
        make_mounts: MountsMaker,
        colour: executor.Colour,
    ) -> executor.Result[_Auth]:
        auth = _Auth(oidc_config, credentials, make_mounts, colour)
        if colour == "async":

            async def _execute() -> _Auth:
                auth._token_endpoint = await executor.aresult(auth._get_token_endpoint())
                await executor.aresult(auth._validate(auth._oidc_config))
                return auth

            return _execute()
        auth._token_endpoint = executor.result(auth._get_token_endpoint())
        executor.result(auth._validate(auth._oidc_config))
        return auth

    def _validate(self, oidc_config: OIDC_CONFIG) -> executor.Result[None]:
        if isinstance(self._credentials, AuthClientPassword):
            # The grant_types_supported field is optional and does not have to be present in the response
            if (
                "grant_types_supported" in oidc_config
                and "password" not in oidc_config["grant_types_supported"]
            ):
                raise AuthenticationFailedError(
                    """The grant_types supported by the third-party authentication service are insufficient. Please add
                    the 'password' grant type."""
                )

            def resp(res: str) -> None:
                if res.startswith("https://login.microsoftonline.com"):
                    raise AuthenticationFailedError(
                        """Microsoft/azure does not recommend to authenticate using username and password and this method is
                        not supported by the python client."""
                    )

            return executor.execute(response_callback=resp, method=self._get_token_endpoint)
        return executor.empty(self.__colour)

    def _get_token_endpoint(self) -> executor.Result[str]:
        if self._token_endpoint is not None:
            return executor.return_(self._token_endpoint, self.__colour)

        def resp(res: httpx.Response) -> str:
            data = _decode_json_response_dict(res, "Get token endpoint")
            assert data is not None
            token_endpoint = data["token_endpoint"]
            assert isinstance(token_endpoint, str)
            return token_endpoint

        if self.__colour == "async":

            async def _execute() -> str:
                mounts: Dict[str, httpx.AsyncBaseTransport] = {}
                for key, mount in self.__make_mounts().items():
                    assert isinstance(mount, httpx.AsyncHTTPTransport)
                    mounts[key] = mount
                async with httpx.AsyncClient(mounts=mounts) as client:
                    return resp(await client.get(self._open_id_config_url))

            return _execute()
        mounts: Dict[str, httpx.BaseTransport] = {}
        for key, mount in self.__make_mounts().items():
            assert isinstance(mount, httpx.BaseTransport)
            mounts[key] = mount
        with httpx.Client(mounts=mounts) as client:
            return resp(client.get(self._open_id_config_url))

    def get_auth_session(self) -> Result:
        if isinstance(self._credentials, AuthBearerToken):
            sessions = self._get_session_auth_bearer_token(self._credentials)
        elif isinstance(self._credentials, AuthClientCredentials):
            sessions = self._get_session_client_credential(self._credentials)
        else:
            assert isinstance(self._credentials, AuthClientPassword)
            sessions = self._get_session_user_pw(self._credentials)
        return sessions

    def _get_session_auth_bearer_token(self, config: AuthBearerToken) -> Result:
        token: Dict[str, Union[str, int]] = {"access_token": config.access_token}
        if config.expires_in is not None:
            token["expires_in"] = config.expires_in
        if config.refresh_token is not None:
            token["refresh_token"] = config.refresh_token

        if "refresh_token" not in token:
            _Warnings.auth_no_refresh_token(config.expires_in)

        if self.__colour == "async":

            async def _execute() -> AsyncOAuth2Client:
                return AsyncOAuth2Client(
                    token=token,
                    token_endpoint=await executor.aresult(self._get_token_endpoint()),
                    client_id=self._client_id,
                    default_timeout=AUTH_DEFAULT_TIMEOUT,
                )

            return _execute()
        return OAuth2Client(
            token=token,
            token_endpoint=executor.result(self._get_token_endpoint()),
            client_id=self._client_id,
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )

    def _get_session_user_pw(self, config: AuthClientPassword) -> Result:
        scope: List[str] = self._default_scopes.copy()
        scope.extend(config.scope_list)
        if self.__colour == "async":

            async def _execute() -> AsyncOAuth2Client:
                session = AsyncOAuth2Client(
                    client_id=self._client_id,
                    token_endpoint=await executor.aresult(self._get_token_endpoint()),
                    grant_type="password",
                    scope=scope,
                    default_timeout=AUTH_DEFAULT_TIMEOUT,
                )
                token: dict = await session.fetch_token(
                    username=config.username, password=config.password
                )
                if "refresh_token" not in token:
                    _Warnings.auth_no_refresh_token(token["expires_in"])

                return session

            return _execute()
        session = OAuth2Client(
            client_id=self._client_id,
            token_endpoint=executor.result(self._get_token_endpoint()),
            grant_type="password",
            scope=scope,
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )
        token: dict = session.fetch_token(username=config.username, password=config.password)
        if "refresh_token" not in token:
            _Warnings.auth_no_refresh_token(token["expires_in"])

        return session

    def __get_common_scopes(self) -> executor.Result[List[str]]:
        def resp(res: str) -> List[str]:
            if res.startswith("https://login.microsoftonline.com"):
                return [self._client_id + "/.default"]
            raise MissingScopeError

        return executor.execute(response_callback=resp, method=self._get_token_endpoint)

    def _get_session_client_credential(self, config: AuthClientCredentials) -> Result:
        scope: List[str] = self._default_scopes.copy()

        if config.scope_list is not None:
            scope.extend(config.scope_list)

        if self.__colour == "async":

            async def _execute() -> AsyncOAuth2Client:
                session = AsyncOAuth2Client(
                    client_id=self._client_id,
                    client_secret=config.client_secret,
                    token_endpoint_auth_method="client_secret_post",
                    scope=(
                        scope
                        if len(scope) > 0
                        else await executor.aresult(self.__get_common_scopes())
                    ),
                    token_endpoint=self._token_endpoint,
                    grant_type="client_credentials",
                    token={"access_token": None, "expires_in": -100},
                    default_timeout=AUTH_DEFAULT_TIMEOUT,
                )
                # explicitly fetch tokens. Otherwise, authlib will do it in the background and we might have race-conditions
                await session.fetch_token()
                return session

            return _execute()
        session = OAuth2Client(
            client_id=self._client_id,
            client_secret=config.client_secret,
            token_endpoint_auth_method="client_secret_post",
            scope=(scope if len(scope) > 0 else executor.result(self.__get_common_scopes())),
            token_endpoint=self._token_endpoint,
            grant_type="client_credentials",
            token={"access_token": None, "expires_in": -100},
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )
        # explicitly fetch tokens. Otherwise, authlib will do it in the background and we might have race-conditions
        session.fetch_token()
        return session



from typing import Union

from weaviate.collections.classes.tenants import Tenant, TenantCreate, TenantUpdate

TenantInputType = Union[str, Tenant]
TenantCreateInputType = Union[str, Tenant, TenantCreate]
TenantUpdateInputType = Union[Tenant, TenantUpdate]
TenantOutputType = Tenant



from typing import Literal, Optional


def multi_vector_schema(quantizer: Optional[Literal["pq", "bq", "sq", "rq"]] = None) -> dict:
    return {
        "class": "Something",
        "invertedIndexConfig": {
            "bm25": {"b": 0.75, "k1": 1.2},
            "cleanupIntervalSeconds": 60,
            "stopwords": {"additions": None, "preset": "en", "removals": None},
        },
        "multiTenancyConfig": {
            "autoTenantActivation": False,
            "autoTenantCreation": False,
            "enabled": False,
        },
        "properties": [
            {
                "dataType": ["text"],
                "indexFilterable": True,
                "indexRangeFilters": False,
                "indexSearchable": True,
                "name": "name",
                "tokenization": "word",
            }
        ],
        "replicationConfig": {"asyncEnabled": False, "factor": 1},
        "shardingConfig": {
            "virtualPerPhysical": 128,
            "desiredCount": 1,
            "actualCount": 1,
            "desiredVirtualCount": 128,
            "actualVirtualCount": 128,
            "key": "_id",
            "strategy": "hash",
            "function": "murmur3",
        },
        "vectorConfig": {
            "boi": {
                "vectorIndexConfig": {
                    "skip": False,
                    "cleanupIntervalSeconds": 300,
                    "maxConnections": 32,
                    "efConstruction": 128,
                    "ef": -1,
                    "dynamicEfMin": 100,
                    "dynamicEfMax": 500,
                    "dynamicEfFactor": 8,
                    "vectorCacheMaxObjects": 1000000000000,
                    "flatSearchCutoff": 40000,
                    "distance": "cosine",
                    "pq": {
                        "enabled": quantizer == "pq",
                        "bitCompression": False,
                        "segments": 0,
                        "centroids": 256,
                        "trainingLimit": 100000,
                        "encoder": {"type": "kmeans", "distribution": "log-normal"},
                    },
                    "bq": {"enabled": quantizer == "bq"},
                    "sq": {
                        "enabled": quantizer == "sq",
                        "trainingLimit": 100000,
                        "rescoreLimit": 20,
                    },
                    "rq": {
                        "enabled": quantizer == "rq",
                        "bits": 8,
                        "rescoreLimit": 20,
                    },
                },
                "vectorIndexType": "hnsw",
                "vectorizer": {"none": {}},
            },
            "yeh": {
                "vectorIndexConfig": {
                    "skip": False,
                    "cleanupIntervalSeconds": 300,
                    "maxConnections": 32,
                    "efConstruction": 128,
                    "ef": -1,
                    "dynamicEfMin": 100,
                    "dynamicEfMax": 500,
                    "dynamicEfFactor": 8,
                    "vectorCacheMaxObjects": 1000000000000,
                    "flatSearchCutoff": 40000,
                    "distance": "cosine",
                    "pq": {
                        "enabled": quantizer == "pq",
                        "bitCompression": False,
                        "segments": 0,
                        "centroids": 256,
                        "trainingLimit": 100000,
                        "encoder": {"type": "kmeans", "distribution": "log-normal"},
                    },
                    "bq": {"enabled": quantizer == "bq"},
                    "sq": {
                        "enabled": quantizer == "sq",
                        "trainingLimit": 100000,
                        "rescoreLimit": 20,
                    },
                    "rq": {
                        "enabled": quantizer == "rq",
                        "bits": 8,
                        "rescoreLimit": 20,
                    },
                },
                "vectorIndexType": "hnsw",
                "vectorizer": {"none": {}},
            },
        },
    }



from typing import Tuple, Union

import pytest

import weaviate
from weaviate.classes.init import AdditionalConfig, Timeout
from weaviate.embedded import WEAVIATE_VERSION
from weaviate.exceptions import WeaviateClosedClientError


@pytest.mark.parametrize("timeout", [(1, 2), Timeout(query=1, insert=2, init=2)])
def test_client_with_extra_options(
    timeout: Union[Tuple[int, int], Timeout], tmp_path_factory: pytest.TempPathFactory
) -> None:
    additional_config = AdditionalConfig(timeout=timeout, trust_env=True)
    client = weaviate.connect_to_embedded(
        port=8070,
        grpc_port=50040,
        additional_config=additional_config,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    )
    try:
        assert client._connection.timeout_config == Timeout(query=1, insert=2, init=2)
    finally:
        client.close()


def test_connect_and_close_to_embedded(tmp_path_factory: pytest.TempPathFactory) -> None:
    # Can't use the default port values as they are already in use by the local instances
    client = weaviate.connect_to_embedded(
        port=30668,
        grpc_port=50151,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    )
    try:
        assert client.is_connected()
        metadata = client.get_meta()
        assert WEAVIATE_VERSION == metadata["version"]
        assert client.is_ready()
        assert "30668" == metadata["hostname"].split(":")[2]
        assert client.is_live()

        client.close()
        assert not client.is_connected()
        with pytest.raises(WeaviateClosedClientError):
            client.get_meta()
    finally:
        client.close()


def test_embedded_as_context_manager(tmp_path_factory: pytest.TempPathFactory) -> None:
    with weaviate.connect_to_embedded(
        port=30668,
        grpc_port=50152,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    ) as client:
        assert client.is_connected()
        metadata = client.get_meta()
        assert WEAVIATE_VERSION == metadata["version"]
        assert client.is_ready()
        assert client.is_live()

    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        client.get_meta()


@pytest.mark.asyncio
async def test_embedded_with_async_as_context_manager(
    tmp_path_factory: pytest.TempPathFactory,
) -> None:
    async with weaviate.use_async_with_embedded(
        port=8076,
        grpc_port=50153,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    ) as client:
        assert client.is_connected()
        metadata = await client.get_meta()
        assert WEAVIATE_VERSION == metadata["version"]
        assert client.is_ready()
        assert client.is_live()

    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        await client.get_meta()


def test_embedded_with_wrong_version(tmp_path_factory: pytest.TempPathFactory) -> None:
    with pytest.raises(weaviate.exceptions.WeaviateEmbeddedInvalidVersionError):
        weaviate.connect_to_embedded(
            version="this_version_does_not_exist",
            environment_variables={"DISABLE_TELEMETRY": "true"},
            persistence_data_path=tmp_path_factory.mktemp("data"),
            binary_path=tmp_path_factory.mktemp("bin"),
        )


def test_embedded_already_running(tmp_path_factory: pytest.TempPathFactory) -> None:
    client = weaviate.connect_to_embedded(
        port=8096,
        grpc_port=50155,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    )
    try:
        assert client._connection.embedded_db is not None
        assert client._connection.embedded_db.process is not None

        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8096, grpc_port=50155, environment_variables={"DISABLE_TELEMETRY": "true"}
            )
    finally:
        client.close()


def test_embedded_startup_with_blocked_http_port(tmp_path_factory: pytest.TempPathFactory) -> None:
    client = weaviate.connect_to_embedded(
        port=8098,
        grpc_port=50096,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    )
    try:
        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8098,
                grpc_port=50097,
                environment_variables={"DISABLE_TELEMETRY": "true"},
                persistence_data_path=tmp_path_factory.mktemp("data"),
                binary_path=tmp_path_factory.mktemp("bin"),
            )
    finally:
        client.close()


def test_embedded_startup_with_blocked_grpc_port(tmp_path_factory: pytest.TempPathFactory) -> None:
    client = weaviate.connect_to_embedded(
        port=8099,
        grpc_port=50150,
        environment_variables={"DISABLE_TELEMETRY": "true"},
        persistence_data_path=tmp_path_factory.mktemp("data"),
        binary_path=tmp_path_factory.mktemp("bin"),
    )
    try:
        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8100, grpc_port=50150, environment_variables={"DISABLE_TELEMETRY": "true"}
            )
    finally:
        client.close()


def test_embedded_create_restore_backup(tmp_path_factory: pytest.TempPathFactory) -> None:
    num_objects = 10
    collection_name = "BackedUp"
    backup_id = "backup0"

    client = weaviate.connect_to_embedded(
        persistence_data_path=tmp_path_factory.mktemp("data"),
        port=8164,
        grpc_port=50500,
        environment_variables={
            "ENABLE_MODULES": "backup-filesystem",
            "BACKUP_FILESYSTEM_PATH": tmp_path_factory.mktemp("backup"),
            "DISABLE_TELEMETRY": "true",
        },
        version="latest",
    )

    try:
        col = client.collections.create(collection_name)
        with col.batch.dynamic() as batch:
            for i in range(num_objects):
                batch.add_object(properties={"text": f"text-{i}"})
        assert len(col) == num_objects

        col.backup.create(backup_id=backup_id, backend="filesystem", wait_for_completion=True)
        client.collections.delete(collection_name)
        col.backup.restore(backup_id=backup_id, backend="filesystem", wait_for_completion=True)

        assert len(col) == num_objects
        client.collections.delete(collection_name)
    finally:
        client.close()



import uuid

from weaviate.collections.classes.batch import MAX_STORED_RESULTS, BatchObjectReturn


def test_batch_object_return_add() -> None:
    lhs_uuids = [uuid.uuid4() for _ in range(MAX_STORED_RESULTS)]
    lhs = BatchObjectReturn(
        _all_responses=lhs_uuids,
        elapsed_seconds=0.1,
        errors={},
        has_errors=False,
        uuids=dict(e for e in enumerate(lhs_uuids)),
    )
    rhs_uuids = [uuid.uuid4() for _ in range(2)]
    rhs = BatchObjectReturn(
        _all_responses=rhs_uuids,
        elapsed_seconds=0.1,
        errors={},
        has_errors=False,
        uuids={
            MAX_STORED_RESULTS: rhs_uuids[0],
            MAX_STORED_RESULTS + 1: rhs_uuids[1],
        },
    )
    result = lhs + rhs
    assert len(result.all_responses) == MAX_STORED_RESULTS
    assert len(result.uuids) == MAX_STORED_RESULTS
    assert result.uuids == {
        idx + len(rhs_uuids): v
        for idx, v in enumerate(lhs_uuids[len(rhs_uuids) : MAX_STORED_RESULTS] + rhs_uuids)
    }



import pytest

from test.collection.schema import multi_vector_schema
from weaviate.collections.classes.config import Reconfigure, _CollectionConfigUpdate
from weaviate.exceptions import WeaviateInvalidInputError


@pytest.mark.parametrize(
    "schema,should_error",
    [
        (multi_vector_schema(), False),
        (multi_vector_schema("bq"), True),
        (multi_vector_schema("sq"), True),
    ],
)
def test_enabling_pq_multi_vector(schema: dict, should_error: bool) -> None:
    update = _CollectionConfigUpdate(
        vectorizer_config=[
            Reconfigure.NamedVectors.update(
                name="boi",
                vector_index_config=Reconfigure.VectorIndex.hnsw(
                    quantizer=Reconfigure.VectorIndex.Quantizer.pq()
                ),
            )
        ]
    )
    if should_error:
        with pytest.raises(WeaviateInvalidInputError):
            update.merge_with_existing(schema)
        return

    new_schema = update.merge_with_existing(schema)

    assert new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["pq"]["enabled"]
    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["bq"]["enabled"]
    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["sq"]["enabled"]

    assert new_schema["vectorConfig"]["yeh"] == schema["vectorConfig"]["yeh"]


@pytest.mark.parametrize(
    "schema,should_error",
    [
        (multi_vector_schema(), False),
        (multi_vector_schema("pq"), True),
        (multi_vector_schema("sq"), True),
    ],
)
def test_enabling_bq_multi_vector(schema: dict, should_error: bool) -> None:
    update = _CollectionConfigUpdate(
        vectorizer_config=[
            Reconfigure.NamedVectors.update(
                name="boi",
                vector_index_config=Reconfigure.VectorIndex.hnsw(
                    quantizer=Reconfigure.VectorIndex.Quantizer.bq()
                ),
            )
        ]
    )
    if should_error:
        with pytest.raises(WeaviateInvalidInputError):
            update.merge_with_existing(schema)
        return

    new_schema = update.merge_with_existing(schema)

    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["pq"]["enabled"]
    assert new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["bq"]["enabled"]
    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["sq"]["enabled"]

    assert new_schema["vectorConfig"]["yeh"] == schema["vectorConfig"]["yeh"]


@pytest.mark.parametrize(
    "schema,should_error",
    [
        (multi_vector_schema(), False),
        (multi_vector_schema("pq"), True),
        (multi_vector_schema("bq"), True),
    ],
)
def test_enabling_sq_multi_vector(schema: dict, should_error: bool) -> None:
    update = _CollectionConfigUpdate(
        vectorizer_config=[
            Reconfigure.NamedVectors.update(
                name="boi",
                vector_index_config=Reconfigure.VectorIndex.hnsw(
                    quantizer=Reconfigure.VectorIndex.Quantizer.sq()
                ),
            )
        ]
    )
    if should_error:
        with pytest.raises(WeaviateInvalidInputError):
            update.merge_with_existing(schema)
        return

    new_schema = update.merge_with_existing(schema)

    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["pq"]["enabled"]
    assert not new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["bq"]["enabled"]
    assert new_schema["vectorConfig"]["boi"]["vectorIndexConfig"]["sq"]["enabled"]

    assert new_schema["vectorConfig"]["yeh"] == schema["vectorConfig"]["yeh"]



import pytest

from integration.conftest import (
    AsyncClientFactory,
    AsyncCollectionFactory,
    ClientFactory,
    CollectionFactory,
)
from weaviate.classes.config import DataType, Property
from weaviate.classes.debug import DebugRESTObject


def test_get_object_single_node(
    client_factory: ClientFactory, collection_factory: CollectionFactory
) -> None:
    client = client_factory()
    collection = collection_factory(properties=[Property(name="name", data_type=DataType.TEXT)])

    uuid = collection.data.insert({"name": "John Doe"})

    debug_obj = client.debug.get_object_over_rest(collection.name, uuid)
    assert debug_obj is not None
    assert isinstance(debug_obj, DebugRESTObject)
    assert str(debug_obj.uuid) == str(uuid)

    non_existant_uuid = "00000000-0000-0000-0000-000000000000"
    debug_obj = client.debug.get_object_over_rest(collection.name, non_existant_uuid)
    assert debug_obj is None


@pytest.mark.asyncio
async def test_get_object_single_node_async(
    async_client_factory: AsyncClientFactory, async_collection_factory: AsyncCollectionFactory
) -> None:
    client = await async_client_factory()
    collection = await async_collection_factory(
        properties=[Property(name="name", data_type=DataType.TEXT)]
    )

    uuid = await collection.data.insert({"name": "John Doe"})

    debug_obj = await client.debug.get_object_over_rest(collection.name, uuid)
    assert debug_obj is not None
    assert isinstance(debug_obj, DebugRESTObject)
    assert str(debug_obj.uuid) == str(uuid)

    non_existant_uuid = "00000000-0000-0000-0000-000000000000"
    debug_obj = await client.debug.get_object_over_rest(collection.name, non_existant_uuid)
    assert debug_obj is None


def test_get_object_multi_node(
    client_factory: ClientFactory, collection_factory: CollectionFactory
) -> None:
    client = client_factory(ports=(8087, 50058))
    collection = collection_factory(
        ports=(8087, 50058), properties=[Property(name="name", data_type=DataType.TEXT)]
    )

    uuid = collection.data.insert({"name": "John Doe"})

    for node_name in ["node1", "node2", "node3"]:
        debug_obj = client.debug.get_object_over_rest(collection.name, uuid, node_name=node_name)
        assert debug_obj is not None
        assert str(debug_obj.uuid) == str(uuid)



from weaviate.collections.classes.batch import (
    DeleteManyObject,
    DeleteManyReturn,
)
from weaviate.collections.classes.data import Error, RefError

__all__ = [
    "DeleteManyObject",
    "DeleteManyReturn",
    "Error",
    "RefError",
]



import uuid
from typing import List, Literal, Optional, Union, overload

from weaviate.cluster.models import ReplicationType, ShardingState
from weaviate.cluster.replicate import _Replicate
from weaviate.cluster.types import Verbosity
from weaviate.collections.classes.cluster import NodeMinimal, NodeVerbose
from weaviate.connect.v4 import ConnectionSync

from .base import _ClusterExecutor

class _Cluster(_ClusterExecutor[ConnectionSync]):
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        shard: Optional[str] = None,
        *,
        output: Literal[None] = None,
    ) -> List[NodeMinimal]: ...
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        shard: Optional[str] = None,
        *,
        output: Literal["minimal"],
    ) -> List[NodeMinimal]: ...
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        shard: Optional[str] = None,
        *,
        output: Literal["verbose"],
    ) -> List[NodeVerbose]: ...
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        shard: Optional[str] = None,
        *,
        output: Optional[Verbosity] = None,
    ) -> Union[List[NodeMinimal], List[NodeVerbose]]: ...
    def replicate(
        self,
        *,
        collection: str,
        shard: str,
        source_node: str,
        target_node: str,
        replication_type: ReplicationType = ReplicationType.COPY,
    ) -> uuid.UUID: ...
    def query_sharding_state(
        self,
        *,
        collection: str,
        shard: Optional[str] = None,
    ) -> Optional[ShardingState]: ...
    @property
    def replications(self) -> _Replicate:
        """replication (_Replication): Replication object instance connected to the same Weaviate instance as the Client.
        This namespace contains all functionality to manage replication operations in Weaviate."""
        ...
